import copy
import inspect
import os
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum, EnumMeta
from functools import wraps
from collections.abc import Mapping
from typing import List, TypeVar, Union, Dict

from ibind.support.logs import project_logger

UNDEFINED = object()
_PRECISION_OFFSET = 7

S = TypeVar('S')
OneOrMany = Union[S, List[S]]

_LOGGER = project_logger(__file__)


def ensure_list_arg(*arg_names: str) -> callable:  # pragma: no cover
    """
    Decorator to ensure that the arguments specified by arg_names are lists.
    If an argument is not a list, it is wrapped in one.
    This version supports ensuring multiple argument names.
    """

    def decorator(func: callable):
        @wraps(func)
        def ensure_list_args_wrapper(*args, **kwargs):
            param_names = list(inspect.signature(func).parameters.keys())

            # Convert args to a list to potentially modify positional arguments
            args_list = list(args)

            for arg_name in arg_names:
                arg_index = param_names.index(arg_name) if arg_name in param_names else None

                # If arg_name was passed as a positional argument
                if arg_index is not None and arg_index < len(args_list):
                    if not isinstance(args_list[arg_index], list):
                        args_list[arg_index] = [args_list[arg_index]]

                # If arg_name was passed as a keyword argument
                elif arg_name in kwargs and not isinstance(kwargs[arg_name], list):
                    kwargs[arg_name] = [kwargs[arg_name]]

            return func(*args_list, **kwargs)

        return ensure_list_args_wrapper

    return decorator


class VerboseEnumMeta(EnumMeta):  # pragma: no cover
    def __getitem__(cls, key):
        return cls.from_string(key)

    def from_string(cls, key):
        enums = list(cls)

        lookup = key.upper().strip()

        for enum in enums:
            if str(enum) == lookup:
                return enum

        raise AttributeError(f'Invalid {cls.__name__}: {key!r} ({type(key)}) | expected: {cls.values()}')

    def values(cls):
        return [entry.value for entry in list(cls)]


class VerboseEnum(str, Enum, metaclass=VerboseEnumMeta):  # pragma: no cover
    """
    A custom base class for enumeration, extending the capabilities of standard Enum types.

    This base class allows for additional functionalities in Enums such as retrieving an enum instance
    using a string representation of its name, and providing a method to get all values of the enum.

    """

    def __init__(self, *args, **kwargs):
        super().__init__()

    def __str__(self):
        return self.value

    def __repr__(self):
        return f'{self.__class__.__name__}.{self.value}'

    def __reduce_ex__(self, proto):
        return self.__class__, (self.value,)

    def __lt__(self, other):
        return self.value < other.value

    def toJSON(self):
        return f'{self.__class__.__name__}.{str(self)}'

    def copy(self):
        return copy.copy(self)


def execute_with_key(key, func, *args, **kwargs):  # pragma: no cover
    try:
        return key, func(*args, **kwargs)
    except Exception as e:
        return key, e


def execute_in_parallel(func: callable, requests: Union[List[dict], Dict[str, dict]], max_workers: int = None, max_per_second: int = 20) -> Union[dict, list]:
    """
    Executes a function in parallel using multiple sets of arguments with rate limiting.


    This function utilises a thread pool to execute the given 'func' concurrently across different sets
    of arguments specified in 'requests'. The 'requests' can be either a list or a dictionary.

    Parameters:
        func (callable): The function to be executed in parallel.
        requests (dict[str, dict] or list): A dictionary where keys are unique identifiers and values are
            dictionaries with 'args' and 'kwargs' for the 'func', or a list of such dictionaries.
        max_workers (int, optional): The maximum number of threads to use.
        max_per_second (int, optional): The maximum number of function executions per second. Defaults to 20.


    Returns:
        Union[dict, list]: A collection of results from the function executions, keyed by the same keys as
        'requests' if it is a dictionary, or a list in the same order as the 'requests' list.
        The function returns results in a dictionary if 'requests' was a dictionary, and a list if  'requests' was a list.
    """

    _requests = requests
    if isinstance(requests, list):
        _requests = {key: request for key, request in enumerate(requests)}

    results = {}
    start_time = time.time()
    num_requests = 0

    with ThreadPoolExecutor(max_workers=max_workers, thread_name_prefix=func.__name__) as executor:
        futures = []
        for key, request in _requests.items():
            while num_requests >= max_per_second:
                if time.time() - start_time >= 1:
                    # Reset the counter and timer every second
                    num_requests = 0
                    start_time = time.time()

            args = request.get('args', [])
            kwargs = request.get('kwargs', {})
            future = executor.submit(execute_with_key, key, func, *args, **kwargs)
            futures.append(future)
            num_requests += 1

        for future in as_completed(futures):
            result = future.result()
            results[result[0]] = result[1]

    if isinstance(requests, list):
        results = [results[key] for key in range(len(requests))]

    return results


def filter_none(d):  # pragma: no cover
    """
    Recursively filters out None values from a dictionary or mapping object.

    This function iteratively processes each key-value pair in the provided dictionary (or any object
    that implements the Mapping interface) and removes any pairs where the value is None. If the value
    itself is a Mapping, the function is applied recursively to it.

    Parameters:
        d (Mapping): The dictionary or mapping object from which to filter out None values.

    Returns:
        Mapping: A new dictionary or mapping object with all None values removed. If the input is not
        a Mapping, it is returned unchanged.

    Note:
        - The function operates recursively, so nested dictionaries will also have None values removed.
        - This function does not modify the original dictionary but returns a new one with the changes.
    """
    if isinstance(d, Mapping):
        return {k: filter_none(v) for k, v in d.items() if v is not None}
    else:
        return d


class TimeoutLock():  # pragma: no cover
    """
    A lock with a timeout mechanism, extending the standard threading.RLock.

    This class provides a reentrant lock (RLock) that can be acquired for a limited duration. If the lock
    cannot be acquired within the specified timeout, the acquire method fails, preventing indefinite blocking.

    Constructor Parameters:
        timeout (int): The maximum time in seconds to wait for the lock to become available.
    """

    def __init__(self, timeout: int):
        self._lock = threading.RLock()
        self._timeout = timeout
        self._acquired = False

    def acquire(self, *args, **kwargs):
        self._acquired = self._lock.acquire(*args, timeout=self._timeout, **kwargs)

    def release(self):
        if self._acquired:
            self._lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, traceback):
        self.release()


# from https://stackoverflow.com/a/37135014/3508719
def exception_to_string(excp) -> str:  # pragma: no cover
    """
    Converts an exception into its string representation including chained exceptions.

    This function takes an exception object as input and returns a string representation of it,
    including any chained exceptions (exceptions raised with 'from' keyword).
    It is typically used for logging purposes or displaying error messages in a user-readable format.

    Parameters:
        excp (Exception): The exception object to be converted into a string.

    Returns:
        str: A string representation of the exception including chained exceptions.
    """
    stack = make_clean_stack() + traceback.extract_tb(excp.__traceback__)
    pretty = traceback.format_list(stack)
    excp_str = '\n'+''.join(pretty) + '\n  {} {}'.format(excp.__class__, excp)

    # Handling chained exceptions
    cause = excp.__cause__
    while cause:
        cause_stack = traceback.extract_tb(cause.__traceback__)
        pretty_cause = traceback.format_list(cause_stack)
        excp_str += '\n\nThe below exception was the direct cause of the above exception:\n\n'
        excp_str += ''.join(pretty_cause) + '\n  {} {}'.format(cause.__class__, cause)
        cause = cause.__cause__

    return excp_str


def make_clean_stack() -> [traceback.FrameSummary]:  # pragma: no cover
    return [s for s in traceback.extract_stack() if all(substring not in s.filename for substring in [
        'JetBrains',
        os.path.join('Lib', 'unittest'),
        os.path.join('Lib', 'logging')])
            ][:-2]


def wait_until(condition: callable, timeout_message: str = None, timeout: float = 5) -> bool:
    """
    Pauses program execution until a specified condition becomes True or a timeout is reached.

   Parameters:
        condition (callable): A callable that returns a boolean value. The function waits until this callable returns True.
        timeout_message (str, optional): A message to log as an error if the timeout is reached. If None, no message is logged. Defaults to None.
        timeout (float, optional): The maximum time to wait for the condition to become True, in seconds. Defaults to 5 seconds.

    Returns:
        bool: True if the condition becomes True within the timeout period, False otherwise.
    """

    deadline = time.time() + timeout
    while time.time() < deadline:
        if condition():
            return True
        time.sleep(0.1)

    if timeout_message is not None:
        _LOGGER.error(timeout_message)

    return False



def tname():  # pragma: no cover
    """
    Generates a unique name for the current thread.

    Returns:
        str: A string combining the current thread's name and its unique identifier.
    """
    return f'{threading.current_thread().name}-{threading.get_ident()}'



def params_dict(required: dict = None, optional: dict = None, preprocessors: dict = None):
    d = required if required is not None else {}

    if optional is None:
        return d

    for key, value in optional.items():
        if value is not None and value != [None]:
            if preprocessors is not None and key in preprocessors:
                value = preprocessors[key](value)
            d[key] = value

    if d == {}:
        return None

    return d

def print_table(my_dict, column_order=None):
    if not column_order:
        column_order = list(my_dict[0].keys() if my_dict else [])
    rv = [column_order] # 1st row = header
    for item in my_dict:
        rv.append([str(item[col] if item[col] is not None else '') for col in column_order])
    column_size = [max(map(len,col)) for col in zip(*rv)]
    formatter = '   '.join(["{{:>{}}}".format(i) for i in column_size])
    for i, item in enumerate(rv):
        print(formatter.format(*item))


from typing import TYPE_CHECKING

from requests import ConnectTimeout

from ibind.base.rest_client import Result
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient

_LOGGER = project_logger(__file__)


class SessionMixin():
    """
    https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#session
    """

    def authentication_status(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        Current Authentication status to the Brokerage system. Market Data and Trading is not possible if not authenticated, e.g. authenticated shows false.
        """
        return self.post('iserver/auth/status')

    def initialize_brokerage_session(self: 'IbkrClient', publish: bool, compete: bool) -> Result:  # pragma: no cover
        """
        After retrieving the access token and subsequent Live Session Token, customers can initialize their brokerage session with the ssodh/init endpoint.
        NOTE: This is essential for using all /iserver endpoints, including access to trading and market data.

        Parameters:
            publish (Boolean): Determines if the request should be sent immediately. Users should always pass true. Otherwise, a ‘500’ response will be returned.
            compete (Boolean): Determines if other brokerage sessions should be disconnected to prioritize this connection.
        """
        return self.post('iserver/auth/ssodh/init', {'publish': publish, 'compete': compete})

    def logout(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        Logs the user out of the gateway session. Any further activity requires re-authentication.
        """
        return self.post('logout')

    def tickle(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        If the gateway has not received any requests for several minutes an open session will automatically timeout. The tickle endpoint pings the server to prevent the session from ending. It is expected to call this endpoint approximately every 60 seconds to maintain the connection to the brokerage session.
        """
        return self.post('tickle', log=False)

    def reauthenticate(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        When using the CP Gateway, this endpoint provides a way to reauthenticate to the Brokerage system as long as there is a valid brokerage session.
        All interest in reauthenticating the gateway session should be handled using the /iserver/auth/ssodh/init endpoint.
        """
        return self.post('iserver/reauthenticate')

    def validate(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        Validates the current session for the SSO user.
        """
        return self.get('/sso/validate')

    def check_health(self: 'IbkrClient') -> bool:
        """
        Verifies the health and authentication status of the IBKR Gateway server.

        This method checks if the Gateway server is alive and whether the user is authenticated.
        It also checks for any competing connections and the connection status.

        Returns:
            bool: True if the Gateway server is authenticated, not competing, and connected, False otherwise.

        Raises:
            AttributeError: If the Gateway health check request returns invalid data.

        Note:
            - This method returns a boolean directly without the `Result` dataclass.
        """
        try:
            result = self.tickle()
        except Exception as e:
            if isinstance(e, ExternalBrokerError) and e.status_code == 401:
                _LOGGER.info(f'Gateway session is not authenticated.')
            elif isinstance(e, ConnectTimeout):
                _LOGGER.error(f'ConnectTimeout raised when communicating with the Gateway. This could indicate that the Gateway is not running or other connectivity issues.')
            else:
                _LOGGER.error(f'Tickle request failed: {e}')
            return False

        if result.data.get('iserver', {}).get('authStatus', {}).get('authenticated', None) is None:
            raise AttributeError(f'Health check requests returns invalid data: {result}')

        auth_status = result.data['iserver']['authStatus']

        authenticated = auth_status['authenticated']
        competing = auth_status['competing']
        connected = auth_status['connected']

        return authenticated and (not competing) and connected



import json
import logging
import time
from threading import Thread
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock, patch, call

import requests

from ibind import Result
from ibind.client.ibkr_client import IbkrClient
from ibind.client.ibkr_ws_client import IbkrWsClient, IbkrSubscriptionProcessor, IbkrWsKey
from test.integration.base.websocketapp_mock import create_wsa_mock, init_wsa_mock
from ibind.support.logs import project_logger
from test_utils import RaiseLogsContext, SafeAssertLogs


class TestPreprocessRawMessage(TestCase):

    def setUp(self):
        self.url = 'wss://localhost:5000/v1/api/ws'

        self.ws_client = IbkrWsClient(
            url=self.url,
            ibkr_client=None,
            account_id=None,
            SubscriptionProcessorClass=lambda: None,
        )

    def test_preprocess_with_well_formed_message(self):
        raw_message = json.dumps({'topic': 'actABC', 'args': {'key': 'value'}})
        expected_result = (
            {'topic': 'actABC', 'args': {'key': 'value'}},  # message
            'actABC',  # topic
            {'key': 'value'},  # data
            'a',  # subscribed
            'ctABC'  # channel
        )
        self.assertEqual(self.ws_client._preprocess_raw_message(raw_message), expected_result)


    def test_preprocess_with_unsubscribed_message(self):
        raw_message = json.dumps({'message': 'Unsubscribed'})
        expected_result = ({'message': 'Unsubscribed'}, None, None, None, None)
        self.assertEqual(self.ws_client._preprocess_raw_message(raw_message), expected_result)


class TestIbkrWsClient(TestCase):
    # Assuming IbkrWsClient is the class containing preprocess_raw_message

    def setUp(self):
        # Assuming similar initialization parameters as in WsClient
        self.url = 'wss://localhost:5000/v1/api/ws'
        self.max_reconnect_attempts = 4
        self.max_ping_interval = 38

        self.url_rest = 'https://localhost:5000'
        self.account_id = 'TEST_ACCOUNT_ID'
        self.timeout = 8
        self.max_retries = 4
        self.subscription_retries = 3
        self.client = MagicMock(spec=IbkrClient(
            url=self.url_rest,
            account_id=self.account_id,
            timeout=self.timeout,
            max_retries=self.max_retries,
        ))

        self.SubscriptionProcessorClass = IbkrSubscriptionProcessor

        # Initialize the IbkrWsClient
        self.ws_client = IbkrWsClient(
            url=self.url,
            ibkr_client=self.client,
            account_id=self.account_id,
            SubscriptionProcessorClass=self.SubscriptionProcessorClass,
            subscription_retries=self.subscription_retries,
            subscription_timeout=0.01,
            cacert=False,
            timeout=0.01,
            max_connection_attempts=self.max_reconnect_attempts,
            max_ping_interval=self.max_ping_interval
        )

        self.wsa_mock = create_wsa_mock()
        self.thread_mock = MagicMock(spec=Thread)
        self.thread_mock.start.side_effect = lambda: self.ws_client._run_websocket(self.wsa_mock)

        self.conid = 265598
        self.update_time = 5678765456

    def run_in_test_context(self, fn, expected_errors: list[str] = None, expect_logs: bool = True):
        with patch('ibind.base.ws_client.WebSocketApp', side_effect=lambda *args, **kwargs: init_wsa_mock(self.wsa_mock, *args, **kwargs)), \
                patch('ibind.base.ws_client.Thread', return_value=self.thread_mock) as new_thread_mock, \
                SafeAssertLogs(self, 'ibind', level='DEBUG', logger_level='DEBUG', no_logs=not expect_logs) as cm, \
                RaiseLogsContext(self, 'ibind', level='WARNING', expected_errors=expected_errors) as cm2:

            ws_client_logger = project_logger('ws_client')
            old_level = ws_client_logger.getEffectiveLevel()
            ws_client_logger.setLevel(logging.WARNING)

            self.new_thread_mock = new_thread_mock
            try:
                rv = fn()
            except:
                raise
            finally:
                ws_client_logger.setLevel(old_level)

        return cm, rv

    def _send_payload(self, payload: dict, expected_errors: list[str] = None, expect_logs: bool = True):
        def run():
            success = self.ws_client.start()
            raw_payload = json.dumps(payload)
            self.ws_client.send(raw_payload)
            self.ws_client.shutdown()
            return success

        return self.run_in_test_context(run, expected_errors=expected_errors, expect_logs=expect_logs)

    def _subscribe(self, request: dict, response: Optional[dict], expected_errors: list[str] = None, expect_logs: bool = True):
        def run():
            def override_on_message(wsa_mock: MagicMock, message: str):
                if response is None:
                    return
                raw_message = json.dumps(response)
                wsa_mock._on_message(wsa_mock, raw_message)

            self.ws_client.start()
            self.wsa_mock.on_message.side_effect = override_on_message
            rv = self.ws_client.subscribe(**{'channel': request.get('channel'), 'data': request.get('data'), 'needs_confirmation':request.get('needs_confirmation')})
            self.ws_client.unsubscribe(**{'channel': request.get('channel'), 'data': request.get('data'), 'needs_confirmation': request.get('confirms_unsubscription')})
            self.ws_client.shutdown()
            return rv

        return self.run_in_test_context(run, expected_errors=expected_errors, expect_logs=expect_logs)

    def test_on_message_system_heartbeat(self):
        hb = 12345678
        cm, success = self._send_payload({'topic': 'system', 'hb': hb}, expect_logs=False)
        # print("\n".join([r.msg for r in cm.records]))
        self.assertEqual(self.ws_client._last_heartbeat, hb)

    def test_on_message_act_account_mismatch(self):
        message_data = {'topic': 'act', 'args': {'accounts': ['OTHER_ACCOUNT_ID']}}
        expected_errors = [
            "IbkrWsClient: Account ID mismatch: expected=TEST_ACCOUNT_ID, received=['OTHER_ACCOUNT_ID']"
        ]

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_blt(self):
        bulletin_message = {'topic': 'blt', 'args': {'bulletin_key': 'some_info'}}

        with patch.object(self.ws_client, '_handle_bulletin', MagicMock()) as mock_handle_bulletin:
            cm, success = self._send_payload(bulletin_message, expect_logs=False)
            mock_handle_bulletin.assert_called_once_with(bulletin_message)

    def test_on_message_sts_unauthenticated(self):
        message_data = {'topic': 'sts', 'args': {'authenticated': False}}
        session_id = 6545676

        expected_errors = [
            "IbkrWsClient: Status unauthenticated: {'authenticated': False}",
            f"IbkrWsClient: Unrecognised message without a topic: {{'session': {session_id}}}"  # in real scenario this doesn't appear but due to the way the tests are written, this is the expected behaviour
        ]

        response_mock = MagicMock(spec=requests.Response)
        response_mock.status_code = 200
        response_mock.json.return_value = {'session': session_id, 'data_to_be_ignored': '1234'}

        self.client.tickle.return_value = Result(data=response_mock.json.return_value)

        with patch('ibind.base.rest_client.requests') as requests_mock:
            requests_mock.request.return_value = response_mock
            cm, success = self._send_payload(message_data, expected_errors=expected_errors)

        self.assertEqual(expected_errors, [r.msg for r in cm.records])
        self.wsa_mock.send.assert_called_with(f'{{"session": {session_id}}}')

    def test_on_message_sts_authenticated(self):
        message_data = {'topic': 'sts', 'args': {'authenticated': True}}
        cm, success = self._send_payload(message_data, expect_logs=False)
        self.assertTrue(self.ws_client._logged_in)

    def test_on_message_error(self):
        message_data = {'topic': 'error', 'args': {'error_key': 'error_details'}}
        expected_errors = [f"IbkrWsClient: Error message:  {message_data}"]

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_no_topic_handler(self):
        message_data = {'topic': 'unrecognized_topic', 'args': {'some_key': 'some_value'}}
        expected_errors = [f"IbkrWsClient: Topic \"{message_data['topic']}\" unrecognised. Message: {message_data}"]

        cm, success = self._send_payload(message_data, expected_errors=expected_errors)
        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def test_on_message_handled_without_subscription(self):
        message_data = {'topic': 'some_topic', 'args': {'channel': 'XYZ', 'data': 'info'}}
        expected_errors = [f"IbkrWsClient: Handled a channel \"{message_data['topic'][1:]}\" message that is missing a subscription. Message: {message_data}"]

        with patch.object(self.ws_client, '_handle_subscribed_message', return_value=True):
            cm, success = self._send_payload(message_data, expected_errors=expected_errors)

        self.assertEqual(expected_errors, [r.msg for r in cm.records])

    def _logs_subscriptions(self, full_channel, data=None, needs_confirmation_sub: bool = False, needs_confirmation_unsub: bool = True):
        return [
            f'IbkrWsClient: Subscribed: s{full_channel}{"" if data is None else f"+{json.dumps(data)}"}{"" if not needs_confirmation_sub else " without confirmation."}',
            f'IbkrWsClient: Unsubscribed: u{full_channel}+{json.dumps(data if data is not None else {})}{"" if not needs_confirmation_unsub else " without confirmation."}'
        ]

    def test_on_message_market_data_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.MARKET_DATA)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'data': {"fields": ['55', '71', '84', '86', '88', '85', '87', '7295', '7296', '70']}}
        response = {'topic': f's{full_channel}', 'conid': self.conid, '_updated': self.update_time, 55: 'AAPL', 70: '195.34', 71: '193.67', 87: '24.2M', 7295: '194.10', 84: '195.25', 86: '195.26', 88: '3,500', 85: '500', 6508: '&serviceID1=122&serviceID2=123&serviceID3=203&serviceID4=775&serviceID5=204&serviceID6=206&serviceID7=108&serviceID8=109'}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, request["data"]), [r.msg for r in cm.records])

        self.assertEqual({self.conid: {'_updated': self.update_time, 'conid': self.conid, 'topic': f'smd+{self.conid}', 'ask_price': '195.26', 'ask_size': '500', 'bid_price': '195.25', 'bid_size': '3,500', 'high': '195.34', 'low': '193.67', 'open': '194.10', 'service_params': '&serviceID1=122&serviceID2=123&serviceID3=203&serviceID4=775&serviceID5=204&serviceID6=206&serviceID7=108&serviceID8=109', 'symbol': 'AAPL', 'volume': '24.2M'}}, queue.get())

    def test_on_message_market_history_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.MARKET_HISTORY)
        server_id = 87567
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'data': {"period": '1min', "bar": "1min", "outsideRTH": True, "source": "trades", "format": "%o/%c/%h/%l"}, 'confirms_unsubscription':False}
        response = {'topic': f's{full_channel}', 'serverId': server_id, '_updated': self.update_time, 'conid': self.conid, 'foo': 'bar'}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, request["data"]), [r.msg for r in cm.records])

        self.assertEqual(response, queue.get())
        self.assertIn(server_id, self.ws_client.server_ids(IbkrWsKey.MARKET_HISTORY))

    def test_on_message_trade_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.TRADES)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}'}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel), [r.msg for r in cm.records])
        self.assertEqual(response, queue.get())

    def test_on_message_orders_channel_handling(self):
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.ORDERS)

        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}'}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        self.assertTrue(queue.empty(), 'Queue should be empty')

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response)
            self.assertTrue(success)

        self.assertEqual(self._logs_subscriptions(full_channel, None, True, True), [r.msg for r in cm.records])
        self.assertEqual(response, queue.get())

    def test_subscription_without_confirmation(self):
        channel = 'fake'
        full_channel = f'{channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'needs_confirmation': False, 'confirms_unsubscription':False}
        response = None

        expected_errors = [
            f'IbkrWsClient: Channel subscription timeout: s{full_channel} after {self.subscription_retries} attempts.'
        ]

        with patch.object(self.ws_client, 'has_subscription', return_value=True):
            cm, success = self._subscribe(request, response, expected_errors=expected_errors)
            self.assertTrue(success)

        self.assertEqual([f'IbkrWsClient: Subscribed: s{full_channel} without confirmation.',
                          f'IbkrWsClient: Unsubscribed: u{full_channel}+{{}} without confirmation.'], [r.msg for r in cm.records])

    def test_check_health(self):
        start_time = [100]
        has_active_connection_counter = [0]

        # control time
        def fake_time():
            start_time[0] += 100
            return start_time[0]

        # simulate that we don't have ws connection first
        def has_active_connection():
            has_active_connection_counter[0] += 1
            if has_active_connection_counter[0] <= 2:
                return False
            return True

        # prepare a fake subscription
        queue = self.ws_client.new_queue_accessor(IbkrWsKey.TRADES)
        full_channel = f'{queue.key.channel}+{self.conid}'
        request = {'channel': f'{full_channel}', 'data': {'foo': 'bar'}}
        response = {'topic': f's{full_channel}', '_updated': self.update_time, 'conid': self.conid, 'args': [{'foo': 'bar'}]}

        def run():
            # ensures each time WebSocketApp's mock is created, we override its on_message method
            def override_init_wsa_mock(wsa_mock: MagicMock, *args, **kwargs):
                wsa_mock = init_wsa_mock(wsa_mock, *args, **kwargs)
                wsa_mock.on_message.side_effect = lambda wsa_mock, message: wsa_mock._on_message(wsa_mock, json.dumps(response))
                return wsa_mock

            self.ws_client.start()
            self.ws_client.check_health()
            self.wsa_mock.on_message.side_effect = lambda wsa_mock, message: wsa_mock._on_message(wsa_mock, json.dumps(response))

            # create the original subscription
            self.ws_client.subscribe(**request)

            # we simulate that closing the WebSocket doesn't work since we have connectivity issues
            # self.wsa_mock.on_close.side_effect = lambda x, y, z: None

            # override time.time, ignore check_ping and take control of has_active_connection
            with patch('ibind.client.ibkr_ws_client.time') as time_mock, \
                    patch.object(self.ws_client, 'check_ping', return_value=True), \
                    patch('ibind.base.ws_client.WebSocketApp', side_effect=lambda *args, **kwargs: override_init_wsa_mock(self.wsa_mock, *args, **kwargs)), \
                    patch.object(self.ws_client, '_has_active_connection', side_effect=has_active_connection) as has_active_connection_mock:
                time_mock.time.side_effect = fake_time
                self.ws_client._last_heartbeat = self.max_ping_interval * 1000

                # this should try to close the connection, fail to do so, abandon the WebSocketApp's mock,
                # then recreate a new mock and recreate the connections
                self.ws_client.check_health()

                self.assertTrue(self.ws_client.ready())
                self.assertEqual([call()] * 6, has_active_connection_mock.call_args_list)
                self.ws_client.shutdown()

        expected_errors = [
            f'IbkrWsClient: Last IBKR heartbeat happened 162.00 seconds ago, exceeding the max ping interval of {self.max_ping_interval}. Restarting.',
            # 'IbkrWsClient: Hard reset close timeout',
            # f'IbkrWsClient: Abandoning current WebSocketApp that cannot be closed: {self.wsa_mock}'
        ]

        cm, success = self.run_in_test_context(run, expected_errors=expected_errors)

        channel_subscribed_log = f"IbkrWsClient: Subscribed: s{full_channel}+{json.dumps(request['data'])}"

        self.assertEqual(
            [channel_subscribed_log] +
            expected_errors +
            [
                f'IbkrWsClient: Invalidated subscription: {full_channel}',
                f"IbkrWsClient: Recreating 1/1 subscriptions: {{'{full_channel}': {{'status': False, 'data': {request['data']}, 'needs_confirmation': True, 'subscription_processor': None}}}}",
                channel_subscribed_log,
                f'IbkrWsClient: Invalidated subscription: {full_channel}',
            ], [r.msg for r in cm.records])



import datetime
from pprint import pformat
from unittest import TestCase
from unittest.mock import patch, MagicMock

from requests import ConnectTimeout

from ibind.client.ibkr_client import IbkrClient
from ibind.client.ibkr_utils import StockQuery, filter_stocks
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.base.rest_client import Result
from test.integration.client import ibkr_responses
from test_utils import verify_log, SafeAssertLogs, RaiseLogsContext


@patch('ibind.base.rest_client.requests')
class TestIbkrClientI(TestCase):

    def setUp(self):
        self.url = 'https://localhost:5000'
        self.account_id = 'TEST_ACCOUNT_ID'
        self.timeout = 8
        self.max_retries = 4
        self.client = IbkrClient(
            url=self.url,
            account_id=self.account_id,
            timeout=self.timeout,
            max_retries=self.max_retries,
        )

        self.data = {'Test key': 'Test value'}

        self.response = MagicMock()
        self.response.json.return_value = self.data
        self.default_path = '/test/api/route'
        self.default_url = f'{self.url}/{self.default_path}'
        self.result = Result(data=self.data, request={'url': self.default_url})
        self.maxDiff = 9999

    def test_get_conids(self, requests_mock):
        requests_mock.request.return_value = self.response
        self.response.json.return_value = ibkr_responses.responses['stocks']

        queries = [
            StockQuery(symbol='AAPL', contract_conditions={'isUS': False, 'exchange': 'AEQLIT'}, name_match='APPLE'),
            StockQuery(symbol='BBVA', contract_conditions={'exchange': 'NYSE'}),
            StockQuery(symbol='CDN', contract_conditions={'isUS': False}),
            StockQuery(symbol='CFC', contract_conditions={}),
            StockQuery(symbol='GOOG', contract_conditions={'isUS': False, 'exchange': 'MEXI'}, instrument_conditions={'chineseName': 'Alphabet&#x516C;&#x53F8;'}),
            'HUBS',
            StockQuery(symbol='META', name_match='meta ', contract_conditions={'isUS': False, 'exchange': 'MEXI'}, instrument_conditions={}),
            StockQuery(symbol='MSFT', contract_conditions={'exchange': 'NASDAQ'}),
            StockQuery(symbol='SAN', name_match='SANTANDER', contract_conditions={'isUS': True}),
            StockQuery(symbol='SCHW', contract_conditions={'exchange': 'NYSE'}),
            StockQuery(symbol='TEAM', name_match='ATLASSIAN'),
            StockQuery(symbol='INVALID_SYMBOL')
        ]

        with self.assertLogs(project_logger(), level='INFO') as cm:
            rv = self.client.stock_conid_by_symbol(queries, default_filtering=False)

        for symbol, conid in rv.data.items():
            self.assertIn(symbol, ibkr_responses.responses['filtered_conids'])
            self.assertEqual(conid, ibkr_responses.responses['filtered_conids'][symbol])

    def test_get_conids_exception(self, requests_mock):
        requests_mock.request.return_value = self.response
        self.response.json.return_value = ibkr_responses.responses['stocks']

        symbol = 'AAPL'
        query = StockQuery(symbol=symbol, contract_conditions={'isUS': False}, name_match='APPLE')

        instruments = filter_stocks(query, Result(data={symbol: ibkr_responses.responses["stocks"][symbol]}), default_filtering=False).data[symbol]

        with self.assertRaises(RuntimeError) as cm_err:
            rv = self.client.stock_conid_by_symbol(query, default_filtering=False)

        self.maxDiff = None
        self.assertEqual(
            f'Filtering stock "{symbol}" returned 2 instruments and 2 contracts using following query: {query}.\nPlease use filters to ensure that only one instrument and one contract per symbol is selected in order to avoid conid ambiguity.\nBe aware that contracts are filtered as {{"isUS": True}} by default. Set default_filtering=False to prevent this default filtering or specify custom filters. See inline documentation for more details.\nInstruments returned:\n{pformat(instruments)}',
            str(cm_err.exception))


    def test_get_live_orders_no_filters(self, requests_mock):
        self.client.get = MagicMock(return_value=self.result)
        self.client.live_orders()
        self.client.get.assert_called_with('iserver/account/orders', params=None)

    def test_get_live_orders_with_valid_filters(self, requests_mock):
        self.client.get = MagicMock(return_value=self.result)
        filters = ['inactive', 'filled']
        self.client.live_orders(filters=filters)
        self.client.get.assert_called_with('iserver/account/orders', params={'filters': 'inactive,filled'})

    def test_get_live_orders_with_single_filter(self, requests_mock):
        self.client.get = MagicMock(return_value=self.result)
        self.client.live_orders(filters='submitted')
        self.client.get.assert_called_with('iserver/account/orders', params={'filters': 'submitted'})

    def test_get_live_orders_with_incorrect_filter_type(self, requests_mock):
        self.client.get = MagicMock(return_value=self.result)
        with self.assertRaises(TypeError):
            self.client.live_orders(filters=123)  # Non-list, non-string filter
        self.client.get.assert_not_called()


    def _marketdata_request(self, method, url, *args, **kwargs):
        leaf = url.split('/')[-1]
        if leaf == 'stocks':
            return MagicMock(json=lambda: ibkr_responses.responses['stocks'])  # Mock response for get_conids
        elif leaf == 'history':
            conid = kwargs['params']['conid']
            return MagicMock(json=lambda: self._history_by_conid[conid])

    def test_marketdata_history_by_symbols(self, requests_mock):
        # Mocking the requests module for external interaction
        self._history_by_conid = {ibkr_responses.responses['filtered_conids'][key]:value for key, value in ibkr_responses.responses['history'].items()}
        requests_mock.request.side_effect = self._marketdata_request

        queries = [
            StockQuery(symbol='AAPL', contract_conditions={'isUS': False, 'exchange': 'AEQLIT'}, name_match='APPLE'),
            StockQuery(symbol='BBVA', contract_conditions={'exchange': 'NYSE'}),
            StockQuery(symbol='CDN', contract_conditions={'isUS': False}),
            StockQuery(symbol='CFC', contract_conditions={}),
            StockQuery(symbol='GOOG', contract_conditions={'isUS': False, 'exchange': 'MEXI'}, instrument_conditions={'chineseName': 'Alphabet&#x516C;&#x53F8;'}),
            StockQuery(symbol='HUBS'),
            StockQuery(symbol='META', name_match='meta ', contract_conditions={'isUS': False, 'exchange': 'MEXI'}, instrument_conditions={}),
            StockQuery(symbol='MSFT', contract_conditions={'exchange': 'NASDAQ'}),
            StockQuery(symbol='SAN', name_match='SANTANDER', contract_conditions={'isUS': True}),
            StockQuery(symbol='SCHW', contract_conditions={'exchange': 'NYSE'}),
            StockQuery(symbol='TEAM', name_match='ATLASSIAN'),
        ]

        expected_results = {}

        for query in queries:
            data = ibkr_responses.responses['history'][query.symbol]['data'][0]
            output = {
                'conid': ibkr_responses.responses['filtered_conids'][query.symbol],
                'symbol': query.symbol,
                "open": data['o'],
                "high": data['h'],
                "low": data['l'],
                "close": data['c'],
                "volume": data['v'],
                "date": datetime.datetime.fromtimestamp(data['t'] / 1000)
            }
            expected_results[query.symbol] = output

        expected_errors = [f'Market data for CDN is not live: Delayed',
                           'Market data for CFC is not live: Delayed']

        with SafeAssertLogs(self, 'ibind', level='INFO', logger_level='DEBUG', no_logs=False) as cm, \
                RaiseLogsContext(self, 'ibind', level='ERROR', expected_errors=expected_errors):
            results = self.client.marketdata_history_by_symbols(queries)

        verify_log(self, cm, expected_errors)

        # Assertions to verify the correctness of each field in the result
        for symbol, expected in expected_results.items():
            result = results[symbol][-1]
            self.assertIn(symbol, results)
            self.assertAlmostEqual(result['open'], expected['open'])
            self.assertAlmostEqual(result['high'], expected['high'])
            self.assertAlmostEqual(result['low'], expected['low'])
            self.assertAlmostEqual(result['close'], expected['close'])
            self.assertAlmostEqual(result['volume'], expected['volume'])
            self.assertEqual(result['date'], expected['date'])

    def test_check_health_authenticated_and_connected(self, requests_mock):
        response_data = {
            'iserver': {
                'authStatus': {
                    'authenticated': True,
                    'competing': False,
                    'connected': True
                }
            }
        }
        requests_mock.request.return_value = MagicMock(json=lambda: response_data)
        self.client.tickle = MagicMock(return_value=Result(data=response_data, request={'url': self.default_url}))

        health_status = self.client.check_health()
        self.assertTrue(health_status)
        self.client.tickle.assert_called_once()

    def test_check_health_not_authenticated(self, requests_mock):
        response_data = {
            'iserver': {
                'authStatus': {
                    'authenticated': False,
                    'competing': False,
                    'connected': True
                }
            }
        }
        requests_mock.request.return_value = MagicMock(json=lambda: response_data)
        self.client.tickle = MagicMock(return_value=Result(data=response_data, request={'url': self.default_url}))

        health_status = self.client.check_health()
        self.assertFalse(health_status)

    def test_check_health_competing_connection(self, requests_mock):
        response_data = {
            'iserver': {
                'authStatus': {
                    'authenticated': True,
                    'competing': True,
                    'connected': True
                }
            }
        }
        requests_mock.request.return_value = MagicMock(json=lambda: response_data)
        self.client.tickle = MagicMock(return_value=Result(data=response_data, request={'url': self.default_url}))

        health_status = self.client.check_health()
        self.assertFalse(health_status)

    def test_check_health_connection_error(self, requests_mock):
        requests_mock.request.side_effect = ConnectTimeout
        self.client.tickle = MagicMock(side_effect=ConnectTimeout)

        with self.assertLogs(level='ERROR') as cm:
            health_status = self.client.check_health()
        self.assertFalse(health_status)
        self.assertIn('ConnectTimeout raised when communicating with the Gateway', cm.output[0])

    def test_check_health_external_broker_error_unauthenticated(self, requests_mock):
        requests_mock.request.side_effect = ExternalBrokerError(status_code=401)
        self.client.tickle = MagicMock(side_effect=ExternalBrokerError(status_code=401))

        with self.assertLogs(level='INFO') as cm:
            health_status = self.client.check_health()
        self.assertFalse(health_status)
        self.assertIn('Gateway session is not authenticated.', cm.output[0])

    def test_check_health_invalid_data(self, requests_mock):
        response_data = {}  # Invalid data format
        requests_mock.request.return_value = MagicMock(json=lambda: response_data)
        self.client.tickle = MagicMock(return_value=Result(data=response_data, request={'url': self.default_url}))

        with self.assertRaises(AttributeError) as cm:
            self.client.check_health()
        self.assertIn('Health check requests returns invalid data', str(cm.exception))


    def test_marketdata_unsubscribe_success(self, requests_mock):
        conids = [12345, 67890]
        responses = {
            12345: MagicMock(status_code=200),
            67890: MagicMock(status_code=200)
        }
        requests_mock.request.side_effect = lambda method, url, **kwargs: responses[int(url.split('/')[-2])]
        self.client.get = MagicMock(side_effect=lambda url, *args, **kwargs: Result(data={'success': True}, request={'url': url}), __name__='client_get_mock')

        results = self.client.marketdata_unsubscribe(conids)

        for conid, result in results.items():
            self.assertIn(conid, conids)
            self.assertIsInstance(result, Result)
            self.assertTrue(result.data['success'])


    def test_marketdata_unsubscribe_with_error(self, requests_mock):
        conids = [12345, 67890]
        responses = {
            12345: MagicMock(status_code=404),  # Simulate not found error for one conid
            67890: MagicMock(status_code=200)
        }
        requests_mock.request.side_effect = lambda method, url, **kwargs: responses[int(url.split('/')[-2])]
        self.client.get = MagicMock(side_effect=lambda url, *args, **kwargs: Result(data={'success': True}, request={'url': url}) if '67890' in url else ExternalBrokerError(status_code=404), __name__='client_get_mock')

        results = self.client.marketdata_unsubscribe(conids)

        self.assertIn(12345, results)
        self.assertIn(67890, results)
        self.assertTrue(results[67890].data['success'])


    def test_marketdata_unsubscribe_raises_exception_on_failure(self, requests_mock):
        conids = [12345]
        responses = {
            12345: MagicMock(status_code=500),  # Simulate server error
        }
        requests_mock.request.side_effect = lambda method, url, **kwargs: responses[int(url.split('/')[-2])]
        self.client.post = MagicMock(side_effect=lambda url, *args, **kwargs: ExternalBrokerError(status_code=500), __name__='client_get_mock')

        with self.assertRaises(ExternalBrokerError):
            self.client.marketdata_unsubscribe(conids)




import functools
import logging
import sys
import traceback
import types
import unittest
from unittest import TestCase
from unittest._log import _CapturingHandler, _AssertLogsContext

import _testcapi

from ibind import var
from ibind.support.py_utils import make_clean_stack


def raise_from_context(cm, level='WARNING'):
    for record in cm.records:
        if record.levelno >= getattr(logging, level):
            raise RuntimeError(record.message)


def verify_log(test_case:TestCase, cm, expected_messages, comparison:callable= lambda x, y: x == y):
    messages = [record.msg for record in cm.records]
    missing_expected = expected_messages.copy()
    for i, expected_msg in enumerate(expected_messages):
        for msg in messages:
            if comparison(expected_msg, msg):
                missing_expected.remove(expected_msg)
                break

    if missing_expected:
        test_case.fail("Expected log(s) not found:\n\t{}".format('\n\t'.join(missing_expected)))


def verify_log_simple(test_self, cm, expected_messages):
    for i, msg in enumerate(expected_messages):
        test_self.assertEqual(msg, cm.records[i].msg)

def exact_log(test_case, cm, expected_messages):
    test_case.assertEqual(expected_messages, [record.msg for record in cm.records])


class SafeAssertLogs(_AssertLogsContext):
    """
    The self.assertLogs context manager, that sets log level on the handler instead of logger.

    Original docstring:
    A context manager used to implement TestCase.assertLogs().
    """
    def __init__(self, *args, logger_level:str=None, **kwargs):
        if sys.version_info < (3, 10, 0) and 'no_logs' in kwargs:
            del kwargs['no_logs']

        super().__init__(*args, **kwargs)
        self.logger_level = logger_level

    def __enter__(self, include_original_handlers:bool=False):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        handler.setLevel(self.level)  # this is different, was logger.setLevel
        logger.propagate = False
        if self.logger_level is not None:
            logger.setLevel(getattr(logging, self.logger_level))

        if include_original_handlers:
            logger.handlers += self.old_handlers
            logger.propagate = True
        return handler.watcher

def get_logger_children(main_logger) -> list[logging.Logger]:
    """
    Gets child loggers. Added as a support compat for Python version 3.11 and below.
    Source: https://github.com/python/cpython/blob/3.12/Lib/logging/__init__.py#L1831
    """

    def _hierlevel(logger):
        if logger is logger.manager.root:
            return 0
        return 1 + logger.name.count('.')

    d = main_logger.manager.loggerDict
    # exclude PlaceHolders - the last check is to ensure that lower-level
    # descendants aren't returned - if there are placeholders, a logger's
    # parent field might point to a grandparent or ancestor thereof.
    return [item for item in d.values()
            if isinstance(item, logging.Logger) and item.parent is main_logger and
            _hierlevel(item) == 1 + _hierlevel(item.parent)]


class RaiseLogsContext:
    """
    Raises any messages above the level raised by a logger.

    When used in conjunction with self.assertLogs, make sure this context manager is defined after self.assertLogs.
    """
    def __init__(self,
                 test_case:TestCase,
                 logger_name=None,
                 level='ERROR',
                 expected_errors:[str]=None,
                 comparison: callable = lambda x, y: x == y,
                 ):
        self._test_case = test_case
        self._logger_name = logger_name
        self._level = level
        self._level_no = getattr(logging, level)
        if expected_errors is None:
            expected_errors = []
        self._expected_errors = expected_errors
        self._comparison = comparison

    def monkey_patch_log(self, original_method):
        def new_method(msg, *args, **kwargs):
            stack = make_clean_stack()
            if 'extra' not in kwargs:
                kwargs['extra'] = {}
            kwargs['extra']['manual_trace'] = stack
            return original_method(msg, *args, **kwargs)

        return new_method

    def monkey_patch_loggers(self, loggers):
        for logger in loggers:
            if self._level_no <= logging.ERROR:
                logger.__old_error_method__ = logger.error
                logger.error = self.monkey_patch_log(logger.error)
            if self._level_no <= logging.WARNING:
                logger.__old_warning_method__ = logger.warning
                logger.warning = self.monkey_patch_log(logger.warning)

    def restore_loggers(self, loggers):
        for logger in loggers:
            if self._level_no <= logging.ERROR:
                logger.error = logger.__old_error_method__
            if self._level_no <= logging.WARNING:
                logger.warning = logger.__old_warning_method__

    def __enter__(self):
        # Directly mimicking the wrapper function inside raise_logs
        self._logger = logging.getLogger(self._logger_name)
        loggers_to_be_patched = [self._logger] + get_logger_children(self._logger)
        self.monkey_patch_loggers(loggers_to_be_patched)
        self._context_manager = SafeAssertLogs(self._test_case, self._logger, level=self._level, no_logs=False)
        self._watcher = self._context_manager.__enter__(include_original_handlers=True)
        return self._watcher

    def __exit__(self, exc_type, exc_val, exc_tb):
        loggers_to_be_patched = [self._logger] + get_logger_children(self._logger)
        self.restore_loggers(loggers_to_be_patched)
        # self.context_manager.__exit__(exc_type, exc_val, exc_tb)
        if exc_type is not None:
            if (exc_type == AssertionError and str(exc_val) != f'no logs of level {self._level} or higher triggered on {self._logger.name}'):
                return False
            else:
                _testcapi.set_exc_info(exc_type, exc_val, exc_tb.tb_next)
                return False

        if len(self._watcher.records) > 0:
            for record in self._watcher.records:
                found = False
                for expected_error in self._expected_errors:
                    if self._comparison(expected_error, record.msg):
                        found = True
                        break
                if found:
                    continue
                # if record.exc_info is not None:
                #     raise record.exc_info[1].with_traceback(record.exc_info[2])

                if hasattr(record, 'manual_trace'):
                    raise RuntimeError(f'\n' + ''.join(traceback.format_list(record.manual_trace)) + f'Logger {self._logger} logged an unexpected message:\n{record.msg}')

                raise RuntimeError(f'\n...\nFile "{record.pathname}", line {record.lineno} in {record.funcName}\n{record.msg}')
        return True


def raise_logs(level='ERROR', logger_name=None):
    def _wrapper(fn):
        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            with RaiseLogsContext(self, level=level, logger_name=logger_name):
                return fn(self, *args, **kwargs)

        return wrapper

    return _wrapper


def decorate_methods(decorator, starts_with=""):
    class DecorateMethods(type):
        """ Decorate all methods of the class with the decorator provided """

        def __new__(cls, name, bases, attrs, **kwargs):
            exclude = kwargs.get('exclude', [])

            for attr_name, attr_value in attrs.items():

                if isinstance(attr_value, types.FunctionType) and \
                        attr_name.startswith(starts_with) and \
                        attr_name not in exclude and \
                        not hasattr(attr_value, '__exclude_decorator__') and \
                        not attr_name.startswith('__'):
                    attrs[attr_name] = decorator(attr_value)

            return super(DecorateMethods, cls).__new__(cls, name, bases, attrs)

    return DecorateMethods


class TestCaseWithRaiseLogs(unittest.TestCase, metaclass=decorate_methods(raise_logs(logger_name='ibind'), starts_with='test')):
    ...


def exclude_decorator(fn):
    fn.__exclude_decorator__ = True
    return fn



class ExternalBrokerError(Exception):
    """ Something unexpected happened externally """

    def __init__(self, *args, status_code: int = None, **kwargs):
        self.status_code = status_code
        super().__init__(*args)



import datetime
import pprint
import threading
import time
from dataclasses import dataclass, field
from typing import Optional, Dict, Union, TYPE_CHECKING

from ibind.base.rest_client import Result, pass_result
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.support.py_utils import UNDEFINED, ensure_list_arg, VerboseEnum, OneOrMany, exception_to_string

_LOGGER = project_logger(__file__)

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient


@dataclass
class StockQuery():
    """
    A class to encapsulate query parameters for filtering stock data.

    This class is used to define a set of criteria for filtering stocks, which includes the stock symbol,
    name matching pattern, and conditions for instruments and contracts.

    Attributes:
        symbol (str): The stock symbol to query.
        name_match (Optional[str], optional): A string pattern to match against stock names. Optional.
        instrument_conditions (Optional[dict], optional): Key-value pairs representing conditions to apply to
            stock instruments. Each condition is matched exactly against the instrument's attributes.
        contract_conditions (Optional[dict], optional): Key-value pairs representing conditions to apply to
            stock contracts. Each condition is matched exactly against the contract's attributes.
    """
    symbol: str
    name_match: Optional[str] = field(default=None)
    instrument_conditions: Optional[dict] = field(default=None)
    contract_conditions: Optional[dict] = field(default=None)


StockQueries = OneOrMany[Union[StockQuery, str]]


def _filter(data: dict, conditions: dict) -> bool:
    for key, value in conditions.items():
        if data.get(key, UNDEFINED) != value:
            return False

    return True


def process_instruments(
        instruments: dict,
        name_match: str = None,
        instrument_conditions: dict = None,
        contract_conditions: dict = None,
) -> [dict]:
    """
    Filters a list of instruments based on specified name matching and conditions.

    This function processes each instrument in the given list and filters them based on the provided
    name matching pattern, instrument conditions, and contract conditions. Instruments are filtered
    to match the specified conditions exactly.

    Parameters:
        instruments (List[dict]): A list of instrument dictionaries to be filtered.
        name_match (Optional[str], optional): A string pattern for partial matching of instrument names. If specified, only instruments  with names containing this pattern are included.
        instrument_conditions (Optional[dict], optional): Key-value pairs representing exact conditions to match against each instrument's attributes.
        contract_conditions (Optional[dict], optional): Key-value pairs representing exact conditions to match against each contract within an instrument.

    Returns:
        List[dict]: A filtered list of instruments that meet the specified criteria.
    """
    filtered_instruments = []

    for instrument in instruments:
        # look for a partial match of the instrument name if provided
        if name_match is not None and name_match.upper() not in instrument['name'].upper():
            continue

        # look for an exact match of instrument properties if provided
        if instrument_conditions is not None and not _filter(instrument, instrument_conditions):
            continue

        # filter contracts by conditions provided
        if contract_conditions is not None:
            filtered_contracts = list(
                filter(
                    lambda x: _filter(x, contract_conditions),
                    instrument["contracts"],
                )
            )

            # if no contracts are left, we don't need the instrument
            if not len(filtered_contracts):
                continue

            # if all conditions are  met, accept the instrument and its contracts
            instrument = {**instrument, 'contracts': filtered_contracts}

        filtered_instruments.append(instrument)

    return filtered_instruments


@ensure_list_arg('queries')
def filter_stocks(queries: [StockQuery], result: Result, default_filtering: bool = True):
    stocks = {}
    data = result.data
    for q in queries:
        symbol, name_match, instrument_conditions, contract_conditions = process_query(q, default_filtering)

        if symbol not in data or len(data[symbol]) == 0:
            _LOGGER.error(f'Error getting stocks. Could not find valid instruments {symbol} in result: {result}')
            continue

        filtered_instruments = process_instruments(
            data[symbol],
            name_match,
            instrument_conditions,
            contract_conditions
        )

        stocks[symbol] = filtered_instruments

    return pass_result(stocks, result)


def query_to_symbols(queries):
    return ",".join([q if isinstance(q, str) else q.symbol for q in queries])


def process_query(q, default_filtering: bool = True):
    if isinstance(q, str):
        q = StockQuery(symbol=q)

    symbol = q.symbol
    name_match = q.name_match
    instrument_conditions = q.instrument_conditions
    # look for US contracts only by default
    default_contract_filter = {'isUS': True} if default_filtering else None
    contract_conditions = q.contract_conditions
    if contract_conditions is None:
        contract_conditions = default_contract_filter

    return symbol, name_match, instrument_conditions, contract_conditions


class QuestionType(VerboseEnum):
    PRICE_PERCENTAGE_CONSTRAINT = 'price exceeds the Percentage constraint of 3%'
    ORDER_VALUE_LIMIT = 'exceeds the Total Value Limit of'
    MISSING_MARKET_DATA = 'You are submitting an order without market data. We strongly recommend against this as it may result in erroneous and unexpected trades.'
    STOP_ORDER_RISKS = 'You are about to submit a stop order. Please be aware of the various stop order types available and the risks associated with each one.'


Answers = Dict[Union[QuestionType, str], bool]


def find_answer(question: str, answers: Answers):
    """
    Retrieves a predefined answer for a given question based on question types defined in the
    QuestionType enum.

    This function matches the given question against known question types and returns the corresponding
    predefined answer. It uses the 'QuestionType' enum to identify types of questions and their
    associated answers.

    Parameters:
        question (str): The question for which an answer is sought.
        answers (Answers): A dictionary mapping QuestionType enum members to their corresponding boolean answers.

    Returns:
        bool: The predefined answer (boolean) corresponding to the given question.

    Raises:
        ValueError: If no predefined answer is found for the given question.
    """
    for question_type, answer in answers.items():
        if str(question_type) in question:
            return answer

    raise ValueError(f'No answer found for question: "{question}"')


def handle_questions(original_result: Result, answers: Answers, reply_callback: callable) -> Result:
    """
    Handles a series of interactive questions that may arise during a request, especially when submitting orders.

    This method iteratively processes questions contained within the response data. It expects
    each question to be answered affirmatively to proceed. If a question does not receive
    a positive reply or if there are too many questions (more than 10 attempts), a RuntimeError
    is raised.

    Parameters:
        original_result (support.rest_client.Result): The initial result object containing the data that may include questions.
        answers (Answers): A collection of answers to the expected questions.

    Returns:
        support.rest_client.Result: The updated result object after all questions have been successfully answered.

    Raises:
        RuntimeError: If a question does not receive a positive reply or if there are too many questions.

    See:
        QuestionType: for types of answers currently supported.

    Note:
        - The function assumes that each response will contain at most one question.
    """

    result: Result = original_result.copy()

    questions = []  # we store questions in case we need to show them at the end
    for attempt in range(20):
        data = result.data

        if 'error' in data:
            order_tag = original_result.request["json"]["orders"][0].get("cOID")
            error_match = f'Order couldn\'t be submitted: Local order ID={order_tag} is already registered.'
            if error_match in data['error']:
                raise ExternalBrokerError(f'Order couldn\'t be submitted. Order with order_tag/cOID {order_tag!r} is already registered.')

            raise ExternalBrokerError(f'While handling questions an error was returned: {pprint.pformat(data)}')

        if not isinstance(data, list):
            raise ExternalBrokerError(f'While handling questions unknown data was returned: {data!r}. Request: {result.request}')

        first_data = data[0]  # this assumes submitting only one order at a time

        # we interpret messages as questions, absence of which we interpret as the end of questions
        if 'message' not in first_data:
            if len(data) == 1:
                data = data[0]
            return pass_result(data, original_result)

        if len(data) != 1:
            _LOGGER.warning(f'While handling questions multiple orders were returned: {pprint.pformat(data)}')

        messages = first_data['message']

        if len(messages) != 1:
            _LOGGER.warning(f'While handling questions multiple messages were returned: {pprint.pformat(messages)}')

        question = messages[0]
        question = question.strip().replace('\n', '')  # clean up the question
        answer = find_answer(question, answers)
        questions.append({'q': question, 'a': answer})

        if answer:
            # the result to a reply will either contain another question or a confirmation
            result = reply_callback(first_data['id'], True)
        else:
            raise RuntimeError(f'A question was not given a positive reply. Question: "{question}". Answers: \n{pprint.pformat(answers)}\n. Request: {result.request}')

    raise RuntimeError(f'Too many questions: {original_result}: {questions}')


def make_order_request(
        conid: Union[int, str],
        side: str,
        quantity: float,
        order_type: str,
        acct_id: str,

        # optional
        price: float = None,
        conidex: str = None,
        sec_type: str = None,
        coid: str = None,
        parent_id: str = None,
        listing_exchange: str = None,
        is_single_group: bool = None,
        outside_rth: bool = None,
        aux_price: float = None,
        ticker: str = None,
        tif: str = 'GTC',
        trailing_amt: float = None,
        trailing_type: str = None,
        referrer: str = None,
        cash_qty: float = None,
        fx_qty: float = None,
        use_adaptive: bool = None,
        is_ccy_conv: bool = None,
        allocation_method: str = None,
        strategy: str = None,
        strategy_parameters=None,

):  # pragma: no cover
    """
     Create an order request object. Arguments set as None will not be included.

    Parameters:
        conid (int | str): Identifier of the security to trade.
        side (str): Order side, either 'SELL' or 'BUY'.
        quantity (int): Order quantity in number of shares.
        order_type (str): Type of the order (e.g., LMT, MKT, STP).
        price (float): Order limit price, depends on order type.
        coid (str): Customer Order ID, unique for a 24h span.
        acctId (str, optional): Account ID, defaults to the first account if not provided.

        conidex (str, Optional): Concatenated value of contract identifier and exchange.
        sec_type (str, Optional): Concatenated value of contract-identifier and security type.
        parent_id (str, Optional): Used for child orders in bracket orders, must match the parent's cOID.
        listing_exchange (str, Optional, optional): Exchange for order routing, default is "SMART".
        is_single_group (bool, Optional): Set to True for placing single group orders (OCA).
        outside_rth (bool, Optional): Set to True if the order can be executed outside regular trading hours.
        aux_price (float, Optional): Auxiliary price parameter.
        ticker (str, Optional): Underlying symbol for the contract.
        tif (str, Optional): Time-In-Force for the order (e.g., GTC, OPG, DAY, IOC). Default: "GTC".
        trailing_amt (float, Optional): Trailing amount for TRAIL or TRAILLMT orders.
        trailing_type (str, Optional): Trailing type ('amt' or '%') for TRAIL or TRAILLMT orders.
        referrer (str, Optional): Custom order reference.
        cash_qty (float, Optional): Cash Quantity for the order.
        fx_qty (float, Optional): Cash quantity for Currency Conversion Orders.
        use_adaptive (bool, Optional): Set to True to use the Price Management Algo.
        is_ccy_conv (bool, Optional): Set to True for FX conversion orders.
        allocation_method (str, Optional): Allocation method for FA account orders.
        strategy (str, Optional): IB Algo algorithm to use for the order.
        strategy_parameters (dict, Optional): Parameters for the specified IB Algo algorithm.
     """

    order_request = {}

    if conid is not None:
        order_request["conid"] = int(conid)

    if side is not None:
        order_request["side"] = str(side)

    if quantity is not None:
        order_request["quantity"] = int(quantity)

    if order_type is not None:
        order_request["orderType"] = str(order_type)

    if price is not None:
        order_request["price"] = price

    if coid is not None:
        order_request["cOID"] = coid

    # optional

    if acct_id is not None:
        order_request["acctId"] = acct_id

    if conidex is not None:
        order_request["conidex"] = conidex

    if sec_type is not None:
        order_request["secType"] = sec_type

    if parent_id is not None:
        order_request["parentId"] = parent_id

    if listing_exchange is not None:
        order_request["listingExchange"] = listing_exchange

    if is_single_group is not None:
        order_request["isSingleGroup"] = is_single_group

    if outside_rth is not None:
        order_request["outsideRTH"] = outside_rth

    if aux_price is not None:
        order_request["auxPrice"] = aux_price

    if ticker is not None:
        order_request["ticker"] = ticker

    if tif is not None:
        order_request["tif"] = tif

    if trailing_amt is not None:
        order_request["trailingAmt"] = trailing_amt

    if trailing_type is not None:
        order_request["trailingType"] = trailing_type

    if referrer is not None:
        order_request["referrer"] = referrer

    if cash_qty is not None:
        order_request["cashQty"] = cash_qty

    if fx_qty is not None:
        order_request["fxQty"] = fx_qty

    if use_adaptive is not None:
        order_request["useAdaptive"] = use_adaptive

    if is_ccy_conv is not None:
        order_request["isCcyConv"] = is_ccy_conv

    if allocation_method is not None:
        order_request["allocationMethod"] = allocation_method

    if strategy is not None:
        order_request["strategy"] = strategy

    if strategy_parameters is not None:
        order_request["strategyParameters"] = strategy_parameters

    return order_request


def date_from_ibkr(d: str) -> datetime.datetime:
    try:
        return datetime.datetime(int(d[:4]), int(d[4:6]), int(d[6:8]), int(d[8:10]), int(d[10:12]), int(d[12:14]))
    except ValueError:
        raise ValueError(f'Date seems to be missing fields: year={d[0:4]}, month={d[4:6]}, day={d[6:8]}, hour={d[8:10]}, minute={d[10:12]}, second={d[12:14]}')


def extract_conid(data):
    # by default conid should be made available as 'smh+<conid>', let's look for it
    if 'topic' in data and '+' in data['topic']:
        return data['topic'].split('+')[-1]

    # as a backup we try to include the conid in the payload, IBKR seems to send it back to us
    elif 'payload' in data and 'conid' in data['payload']:
        return data['payload']['conid']

    return None


class Tickler():
    """
    Utility class used for maintaining the OAuth connection alive by repeatedly calling the `tickle` method.
    """

    def __init__(self, client: 'IbkrClient', interval: Union[int, float] = 60):
        """
        Initialise the Tickler instance.

        Args:
            client (IbkrClient): The client instance with a `tickle` method.
            interval (int | float): Interval between tickles in seconds. Default is 60 seconds.
        """
        self._client = client
        self._interval = interval
        self._running = True
        self._thread = None

    def _worker(self):
        _LOGGER.info(f'Tickler starts with interval={self._interval} seconds.')
        while self._running:
            try:
                self._client.tickle()
            except KeyboardInterrupt:
                _LOGGER.info('Tickler interrupted')
                break
            except Exception as e:
                _LOGGER.error(f'Tickler error: {exception_to_string(e)}')

            time.sleep(self._interval)

        _LOGGER.info(f'Tickler gracefully stopped.')

    def start(self):
        if self._thread is not None:
            _LOGGER.info('Tickler thread already running. Stop the existing thread first by calling Tickler.stop()')
            return

        self._thread = threading.Thread(target=self._worker)
        self._thread.daemon = True
        self._thread.start()

    def stop(self, timeout=None):
        if self._thread is None:
            return

        self._running = False
        self._thread.join(timeout)



from unittest import TestCase
from unittest.mock import patch, MagicMock

from requests import ReadTimeout, Timeout

from ibind.client.ibkr_client import IbkrClient
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.base.rest_client import Result, RestClient


@patch('ibind.base.rest_client.requests')
class TestIbkrClientI(TestCase):
    def setUp(self):
        self.url = 'https://localhost:5000'
        self.account_id = 'TEST_ACCOUNT_ID'
        self.timeout = 8
        self.max_retries = 4
        self.client = RestClient(
            url=self.url,
            timeout=self.timeout,
            max_retries=self.max_retries,
        )

        self.data = {'Test key': 'Test value'}

        self.response = MagicMock()
        self.response.json.return_value = self.data
        self.default_path = 'test/api/route'
        self.default_url = f'{self.url}/{self.default_path}'
        self.result = Result(data=self.data, request={'url': self.default_url})
        self.maxDiff = 9999

    def test_default_rest(self, requests_mock):
        requests_mock.request.return_value = self.response

        rv = self.client.get(self.default_path)
        self.assertEqual(self.result, rv)
        requests_mock.request.assert_called_with('GET', self.default_url, verify=False, timeout=self.timeout)

        test_post_kwargs = {'field1': 'value1', 'field2': 'value2'}
        test_json = {'json': {**test_post_kwargs}}
        rv = self.client.post(self.default_path, params=test_post_kwargs)
        self.assertEqual(self.result.copy(request={'url': self.default_url, **test_json}), rv)
        requests_mock.request.assert_called_with('POST', self.default_url, verify=False, timeout=self.timeout, **test_json)

        rv = self.client.delete(self.default_path)
        self.assertEqual(self.result, rv)
        requests_mock.request.assert_called_with('DELETE', self.default_url, verify=False, timeout=self.timeout)

    def test_request_retries(self, requests_mock):
        requests_mock.request.side_effect = ReadTimeout()

        with self.assertLogs(project_logger(), level='INFO') as cm, \
                self.assertRaises(TimeoutError) as cm_err:
            rv = self.client.get(self.default_path)

        for i, record in enumerate(cm.records):
            self.assertEqual(f'RestClient: Timeout for GET {self.default_url}, retrying attempt {i + 1}/{self.max_retries}', record.msg)
        self.assertEqual(f"RestClient: Reached max retries ({self.max_retries}) for GET {self.default_url} {{}}", str(cm_err.exception))

    def test_response_raise_timeout(self, requests_mock):
        requests_mock.request.return_value = self.response
        self.response.raise_for_status.side_effect = Timeout()

        with self.assertRaises(ExternalBrokerError) as cm_err:
            rv = self.client.get(self.default_path)

        self.assertEqual(f"RestClient: Timeout error ({self.timeout}S)", str(cm_err.exception))

    def test_response_raise_generic(self, requests_mock):
        requests_mock.request.return_value = self.response
        self.response.status_code = 400
        self.response.reason = 'Test reason'
        self.response.text = 'Test text'

        self.response.raise_for_status.side_effect = ValueError('Test generic error')

        with self.assertRaises(ExternalBrokerError) as cm_err:
            rv = self.client.get(self.default_path)

        self.assertEqual(f"RestClient: response error {self.result.copy(data=None)} :: {self.response.status_code} :: {self.response.reason} :: {self.response.text}", str(cm_err.exception))



import importlib.util
import os
from typing import Union, Optional

from ibind import var
from ibind.base.rest_client import RestClient
from ibind.client.ibkr_client_mixins.accounts_mixin import AccountsMixin
from ibind.client.ibkr_client_mixins.contract_mixin import ContractMixin
from ibind.client.ibkr_client_mixins.marketdata_mixin import MarketdataMixin
from ibind.client.ibkr_client_mixins.order_mixin import OrderMixin
from ibind.client.ibkr_client_mixins.portfolio_mixin import PortfolioMixin
from ibind.client.ibkr_client_mixins.scanner_mixin import ScannerMixin
from ibind.client.ibkr_client_mixins.session_mixin import SessionMixin
from ibind.client.ibkr_client_mixins.watchlist_mixin import WatchlistMixin
from ibind.client.ibkr_utils import Tickler
from ibind.support.logs import new_daily_rotating_file_handler, project_logger

_LOGGER = project_logger(__file__)


class IbkrClient(RestClient, AccountsMixin, ContractMixin, MarketdataMixin, OrderMixin, PortfolioMixin, ScannerMixin, SessionMixin, WatchlistMixin):
    """
    A client class for interfacing with the IBKR API, extending the RestClient class.

    This subclass of RestClient is specifically designed for the IBKR API. It inherits
    the foundational REST API interaction capabilities from RestClient and adds functionalities
    particular to the IBKR API, such as specific endpoint handling.

    The class provides methods to perform various operations with the IBKR API, such as
    fetching stock data, submitting orders, and managing account information.

    See: https://interactivebrokers.github.io/cpwebapi/endpoints

    Note:
        - All endpoint mappings are defined as class mixins, categorised similar to the IBKR REST API documentation. See appropriate mixins for more information. 
    """

    def __init__(
            self,
            account_id: Optional[str] = var.IBIND_ACCOUNT_ID,
            url: str = var.IBIND_REST_URL,
            host: str = 'localhost',
            port: str = '5000',
            base_route: str = '/v1/api/',
            cacert: Union[str, os.PathLike, bool] = var.IBIND_CACERT,
            timeout: float = 10,
            max_retries: int = 3,
            use_oauth: bool = var.IBIND_USE_OAUTH,
    ) -> None:
        """
        Parameters:
            account_id (str): An identifier for the account. Defaults to None.
            url (str): The base URL for the REST API. Defaults to None.
            host (str, optional): Host for the IBKR REST API. Defaults to 'localhost'.
            port (str, optional): Port for the IBKR REST API. Defaults to '5000'
            base_route (str, optional): Base route for the IBKR REST API. Defaults to '/v1/api/'.
            cacert (Union[os.PathLike, bool], optional): Path to the CA certificate file for SSL verification,
                                                         or False to disable SSL verification. Always True when
                                                         use_oauth is True. Defaults to False.
            timeout (float, optional): Timeout in seconds for the API requests. Defaults to 10.
            max_retries (int, optional): Maximum number of retries for failed API requests. Defaults to 3.
            use_oauth (bool, optional): Whether to use OAuth authentication. Defaults to False.
        """

        self._use_oauth = use_oauth

        url = var.IBIND_OAUTH_REST_URL if self._use_oauth else url

        if url is None:
            url = f'https://{host}:{port}{base_route}'

        self.account_id = account_id

        cacert = True if self._use_oauth else cacert
        super().__init__(url=url, cacert=cacert, timeout=timeout, max_retries=max_retries)

        self.logger.info('#################')
        self.logger.info(f'New IbkrClient(base_url={self.base_url!r}, account_id={self.account_id!r}, ssl={self.cacert!r}, timeout={self._timeout}, max_retries={self._max_retries})')

        if self._use_oauth:
            self.oauth_init()

    def make_logger(self):
        self._logger = new_daily_rotating_file_handler('IbkrClient', os.path.join(var.LOGS_DIR, f'ibkr_client_{self.account_id}'))

    def generate_live_session_token(self):
        from ibind.support.oauth import req_live_session_token
        self.live_session_token, self.live_session_token_expires_ms, self.live_session_token_signature = req_live_session_token(self)

    def oauth_init(self):
        if importlib.util.find_spec('Crypto') is None:
            raise ImportError('Installation lacks OAuth support. Please install by using `pip install ibind[oauth]`')

        # get live session token for OAuth authentication
        self.generate_live_session_token()

        # validate the live session token once
        from ibind.support.oauth import validate_live_session_token
        success = validate_live_session_token(
            live_session_token=self.live_session_token,
            live_session_token_signature=self.live_session_token_signature,
            consumer_key=var.IBIND_CONSUMER_KEY
        )
        if not success:
            raise RuntimeError("Live session token validation failed.")

        # start Tickler to maintain the connection alive
        self._tickler = Tickler(self)
        self._tickler.start()

        # add signal handlers to gracefully shut down the Tickler and the client
        import signal
        existing_handler_int = signal.getsignal(signal.SIGINT)
        existing_handler_term = signal.getsignal(signal.SIGTERM)

        def _stop(signum, frame):
            self.oauth_shutdown()

            if signum == signal.SIGINT and callable(existing_handler_int):
                existing_handler_int(signum, frame)

            if signum == signal.SIGTERM and callable(existing_handler_term):
                existing_handler_term(signum, frame)

        signal.signal(signal.SIGINT, _stop)
        signal.signal(signal.SIGTERM, _stop)

    def oauth_shutdown(self):
        _LOGGER.info(f'IbkrClient: Shutting down OAuth')

        if hasattr(self, '_tickler') and self._tickler is not None:
            self._tickler.stop()

        self.logout()

    def get_headers(self, request_method: str, request_url: str):
        if (not self._use_oauth) or request_url == f'{self.base_url}{var.IBIND_LIVE_SESSION_TOKEN_ENDPOINT}':
            # No need for extra headers if we don't use oauth or getting live session token
            return {}

        # get headers for endpoints other than live session token request
        from ibind.support.oauth import generate_oauth_headers
        headers = generate_oauth_headers(
            request_method=request_method,
            request_url=request_url,
            live_session_token=self.live_session_token
        )

        return headers



from typing import TYPE_CHECKING

from ibind.base.rest_client import Result
from ibind.support.logs import project_logger
from ibind.support.py_utils import params_dict, ensure_list_arg, OneOrMany

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient

_LOGGER = project_logger(__file__)


class PortfolioMixin():  # pragma: no cover
    """
    * https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#portfolio
    * https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#pa
    """

    def portfolio_accounts(self: 'IbkrClient') -> Result:
        """
        In non-tiered account structures, returns a list of accounts for which the user can view position and account information. This endpoint must be called prior to calling other /portfolio endpoints for those accounts.
        """
        return self.get('portfolio/accounts')

    def portfolio_subaccounts(self: 'IbkrClient') -> Result:
        """
        Used in tiered account structures (such as Financial Advisor and IBroker Accounts) to return a list of up to 100 sub-accounts for which the user can view position and account-related information. This endpoint must be called prior to calling other /portfolio endpoints for those sub-accounts.
        """
        return self.get('portfolio/subaccounts')

    def large_portfolio_subaccounts(self: 'IbkrClient', page: int = 0) -> Result:
        """
        Used in tiered account structures (such as Financial Advisor and IBroker Accounts) to return a list of sub-accounts, paginated up to 20 accounts per page, for which the user can view position and account-related information. This endpoint must be called prior to calling other /portfolio endpoints for those sub-accounts.
        """
        return self.get('portfolio/subaccounts2', {'page': page})

    def portfolio_account_information(self: 'IbkrClient', account_id: str = None) -> Result:
        """
        Account information related to account Id. /portfolio/accounts or /portfolio/subaccounts must be called prior to this endpoint.

        Parameters:
            account_id (str, optional): Specify the AccountID to receive portfolio information for.
        """
        if account_id == None:
            account_id = self.account_id
        return self.get(f'portfolio/{account_id}/meta')

    def portfolio_account_allocation(self: 'IbkrClient', account_id: str = None) -> Result:
        """
        Information about the account's portfolio allocation by Asset Class, Industry and Category. /portfolio/accounts or /portfolio/subaccounts must be called prior to this endpoint.

        Parameters:
            account_id (str, optional): Specify the account ID for the request.
        """
        if account_id == None:
            account_id = self.account_id
        return self.get(f'portfolio/{account_id}/allocation')

    @ensure_list_arg('account_ids')
    def portfolio_account_allocations(self: 'IbkrClient', account_ids: OneOrMany[str]) -> Result:
        """
        Similar to /portfolio/{accountId}/allocation but returns a consolidated view of all the accounts returned by /portfolio/accounts.

        Parameters:
            account_ids (OneOrMany[str]): Contains all account IDs as strings the user should receive data for.
        """
        params = params_dict({'acctIds': account_ids})
        return self.get(f'portfolio/allocation', params=params)

    def positions(
            self: 'IbkrClient',
            account_id: str = None,
            page: int = 0,
            model: str = None,
            sort: str = None,
            direction: str = None,
            period: str = None,
    ) -> Result:
        """
        Returns a list of positions for the given account. The endpoint supports paging, each page will return up to 100 positions.

        Parameters:
            account_id (str, optional): The account ID for which account should place the order.
            page_id (str, optional): The “page” of positions that should be returned. One page contains a maximum of 100 positions. Pagination starts at 0.
            model (str, optional): Code for the model portfolio to compare against.
            sort (str, optional): Declare the table to be sorted by which column.
            direction (str, optional): The order to sort by. 'a' means ascending 'd' means descending.
            period (str, optional): Period for pnl column. Value Format: 1D, 7D, 1M.
        """

        if account_id == None:
            account_id = self.account_id

        params = params_dict(
            optional={
                'model': model,
                'sort': sort,
                'direction': direction,
                'period': period,
            }
        )
        return self.get(f'portfolio/{account_id}/positions/{page}', params)

    def positions2(
            self: 'IbkrClient',
            account_id: str = None,
            model: str = None,
            sort: str = None,
            direction: str = None,
    ) -> Result:
        """
        Returns a list of positions for the given account.
        /portfolio/accounts or /portfolio/subaccounts must be called prior to this endpoint.
        This endpoint provides near-real time updates and removes caching otherwise found in the /portfolio/{accountId}/positions/{pageId} endpoint.

        Parameters:
            account_id (str, optional): The account ID for which account should place the order.
            model (str, optional): Code for the model portfolio to compare against.
            sort (str, optional): Declare the table to be sorted by which column.
            direction (str, optional): The order to sort by. 'a' means ascending 'd' means descending.
        """

        if account_id == None:
            account_id = self.account_id

        params = params_dict(
            optional={
                'model': model,
                'sort': sort,
                'direction': direction,
            }
        )
        return self.get(f'portfolio2/{account_id}/positions', params)

    def positions_by_conid(self: 'IbkrClient', account_id: str, conid: str) -> Result:
        """
        Returns a list containing position details only for the specified conid.

        Parameters:
            account_id (str): The account ID for which account should place the order.
            conid (str): The contract ID to receive position information on.
        """
        if account_id == None:
            account_id = self.account_id
        return self.get(f'/portfolio/{account_id}/position/{conid}')

    def invalidate_backend_portfolio_cache(self: 'IbkrClient', account_id: str = None) -> Result:
        """
        Invalidates the cached value for your portfolio’s positions and calls the /portfolio/{accountId}/positions/0 endpoint automatically.

        Parameters:
            account_id (str): The account ID for which cache to invalidate.
        """
        if account_id is None:
            account_id = self.account_id
        return self.post(f'portfolio/{account_id}/positions/invalidate')

    def portfolio_summary(self: 'IbkrClient', account_id: str = None) -> Result:
        """
        Information regarding settled cash, cash balances, etc. in the account’s base currency and any other cash balances hold in other currencies. /portfolio/accounts or /portfolio/subaccounts must be called prior to this endpoint. The list of supported currencies is available at https://www.interactivebrokers.com/en/index.php?f=3185.

        Parameters:
            account_id (str): Specify the account ID for which account you require ledger information on.
        """
        if account_id is None:
            account_id = self.account_id
        return self.get(f'portfolio/{account_id}/summary')

    def get_ledger(self: 'IbkrClient', account_id: str = None) -> Result:
        """
        Information regarding settled cash, cash balances, etc. in the account’s base currency and any other cash balances hold in other currencies. /portfolio/accounts or /portfolio/subaccounts must be called prior to this endpoint. The list of supported currencies is available at https://www.interactivebrokers.com/en/index.php?f=3185.

        Parameters:
            account_id (str): Specify the account ID for which account you require ledger information on.
        """
        if account_id is None:
            account_id = self.account_id
        return self.get(f'portfolio/{account_id}/ledger')

    def position_and_contract_info(self: 'IbkrClient', conid: str) -> Result:
        """
        Returns an object containing information about a given position along with its contract details.

        Parameters:
            conid (str): The contract ID to receive position information on.
        """
        return self.get(f'portfolio/positions/{conid}')

    @ensure_list_arg('account_ids')
    def account_performance(self: 'IbkrClient', account_ids: OneOrMany[str], period: str) -> Result:
        """
        Returns the performance (MTM) for the given accounts, if more than one account is passed, the result is consolidated.

        Parameters:
            account_ids (OneOrMany[str]): Include each account ID to receive data for.
            period (str): Specify the period for which the account should be analyzed. Available Values: “1D”, “7D”, “MTD”, “1M”, “YTD”, “1Y”.
        """
        return self.post(f'pa/performance', {'acctIds': account_ids, 'period': period})

    @ensure_list_arg('account_ids', 'conids')
    def transaction_history(
            self: 'IbkrClient',
            account_ids: OneOrMany[str],
            conids: OneOrMany[str],
            currency: str,
            days: str = None
    ) -> Result:
        """
        Transaction history for a given number of conids and accounts. Types of transactions include dividend payments, buy and sell transactions, transfers.

        Parameters:
            account_ids (OneOrMany[str]): Include each account ID to receive data for.
            conids (OneOrMany[str]): Include contract ID to receive data for. Only supports one contract id at a time.
            currency (str): Define the currency to display price amounts with. Defaults to USD.
            days (str, optional): Specify the number of days to receive transaction data for. Defaults to 90 days of transaction history if unspecified.
        """
        params = params_dict(
            {
                'acctIds': account_ids,
                'conids': conids,
                'currency': currency,
            }, optional={'days': days}
        )
        return self.post(f'pa/transactions', params)



responses = {
    "accounts": [
        {
            "PrepaidCrypto-P": False,
            "PrepaidCrypto-Z": False,
            "accountAlias": None,
            "accountId": "DUMMY_ACCOUNT_ID123",
            "accountStatus": 1678852800000,
            "accountTitle": "Geralt of Rivia",
            "accountVan": "DUMMY_ACCOUNT_ID123",
            "brokerageAccess": True,
            "businessType": "INDEPENDENT",
            "clearingStatus": "O",
            "covestor": False,
            "currency": "USD",
            "desc": "DUMMY_ACCOUNT_ID123",
            "displayName": "Geralt of Rivia",
            "faclient": False,
            "ibEntity": "IBLLC-US",
            "id": "DUMMY_ACCOUNT_ID123",
            "noClientTrading": False,
            "parent": {
                "accountId": "",
                "isMChild": False,
                "isMParent": False,
                "isMultiplex": False,
                "mmc": [],
            },
            "trackVirtualFXPortfolio": False,
            "tradingType": "STKNOPT",
            "type": "DEMO",
        }
    ],
    "auth_status": {
        "MAC": "F4:03:43:CD:BD:C0",
        "authenticated": True,
        "competing": False,
        "connected": True,
        "fail": "",
        "hardware_info": "1c4558e8|F4:03:43:CD:BD:C0",
        "message": "",
        "serverInfo": {
            "serverName": "JifC11113",
            "serverVersion": "Build 10.25.0m, Nov 17, 2023 " "11:29:27 AM",
        },
    },
    "conids": {
        "AAPL": 265598,
        "BBVA": 4815,
        "CDN": 453310666,
        "CFC": 42001300,
        "GOOG": 208813720,
        "HUBS": 169544810,
        "META": 107113386,
        "MSFT": 272093,
        "SAN": 12442,
        "SCHW": 11905,
        "TEAM": 589316251,
    },
    'filtered_conids': {
        'AAPL': 532640894,
        'BBVA': 4815,
        'CDN': 453310666,
        'CFC': 42001300,
        'GOOG': 210810667,
        'HUBS': 169544810,
        'META': 114922621,
        'MSFT': 272093,
        'SAN': 12442,
        'SCHW': 11905,
        'TEAM': 589316251
    },
    'history': {
        'AAPL': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.24/1440;',
            'data': [{'c': 189.37,
                      'h': 192.93,
                      'l': 170.12,
                      'o': 171.05,
                      't': 1698796800000,
                      'v': 6942998.04}],
            'high': '19293/6942998.04/1440',
            'low': '17012/6942998.04/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195436',
            'startTime': '20231031-00:00:00',
            'symbol': 'AAPL',
            'text': 'APPLE INC',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1
        },
        'BBVA': {
            'barLength': 2678400,
            'data': [{'c': 9.45,
                      'h': 9.52,
                      'l': 7.94,
                      'o': 8.0,
                      't': 1698796800000,
                      'v': 211434.01}],
            'high': '952/211434.01/1440',
            'low': '794/211434.01/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195431',
            'startTime': '20231031-00:00:00',
            'symbol': 'BBVA',
            'text': 'BANCO BILBAO VIZCAYA-SP ADR',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'CDN': {
            'barLength': 2678400,
            'data': [{'c': 0.32,
                      'h': 0.37,
                      'l': 0.26,
                      'o': 0.37,
                      't': 1698796800000,
                      'v': 83773}],
            'high': '74/83773/1440',
            'low': '52/83773/1440',
            'mdAvailability': 'D',
            'messageVersion': 2,
            'mktDataDelay': 900,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '3',
            'priceFactor': 200,
            'serverId': '195417',
            'startTime': '20231031-00:00:00',
            'symbol': 'CDN',
            'text': 'CDN MAVERICK CAPITAL CORP',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1
        },
        'CFC': {
            'barLength': 2678400,
            'data': [{'c': 1.62,
                      'h': 1.91,
                      'l': 1.48,
                      'o': 1.63,
                      't': 1698796800000,
                      'v': 949972}],
            'high': '3820/949972/1440',
            'low': '2960/949972/1440',
            'mdAvailability': 'D',
            'messageVersion': 2,
            'mktDataDelay': 900,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '4',
            'priceFactor': 2000,
            'serverId': '195419',
            'startTime': '20231031-00:00:00',
            'symbol': 'CFC',
            'text': 'UET UNITED ELECTRONIC TECHNO',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'GOOG': {
            'barLength': 2678400,
            'data': [{'c': 136.4,
                      'h': 141.1,
                      'l': 124.92,
                      'o': 125.24,
                      't': 1698796800000,
                      'v': 2346378.23}],
            'high': '14110/2346378.23/1440',
            'low': '12492/2346378.23/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195424',
            'startTime': '20231031-00:00:00',
            'symbol': 'GOOG',
            'text': 'ALPHABET INC-CL C',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'HUBS': {
            'barLength': 2678400,
            'data': [{'c': 496.52,
                      'h': 499.15,
                      'l': 408.62,
                      'o': 422.06,
                      't': 1698796800000,
                      'v': 77847.62}],
            'high': '49915/77847.62/1440',
            'low': '40862/77847.62/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195440',
            'startTime': '20231031-00:00:00',
            'symbol': 'HUBS',
            'text': 'HUBSPOT INC',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1
        },
        'META': {
            'barLength': 2678400,
            'data': [{'c': 332.2,
                      'h': 342.92,
                      'l': 301.85,
                      'o': 301.85,
                      't': 1698796800000,
                      'v': 1826033.84}],
            'high': '34292/1826033.84/1440',
            'low': '30185/1826033.84/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195438',
            'startTime': '20231031-00:00:00',
            'symbol': 'META',
            'text': 'META PLATFORMS INC-CLASS A',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'MSFT': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.75/1440;',
            'data': [{'c': 378.85,
                      'h': 384.3,
                      'l': 339.65,
                      'o': 339.79,
                      't': 1698796800000,
                      'v': 2886663.6}],
            'high': '38430/2886663.6/1440',
            'low': '33965/2886663.6/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195433',
            'startTime': '20231031-00:00:00',
            'symbol': 'MSFT',
            'text': 'MICROSOFT CORP',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'SAN': {
            'barLength': 2678400,
            'data': [{'c': 4.13,
                      'h': 4.15,
                      'l': 3.62,
                      'o': 3.66,
                      't': 1698796800000,
                      'v': 466540.33}],
            'high': '41500/466540.33/1440',
            'low': '36200/466540.33/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '4',
            'priceFactor': 10000,
            'serverId': '195429',
            'startTime': '20231031-00:00:00',
            'symbol': 'SAN',
            'text': 'BANCO SANTANDER SA-SPON ADR',
            'timePeriod': '1m',
            'travelTime': 3,
            'volumeFactor': 1
        },
        'SCHW': {
            'barLength': 2678400,
            'chartAnnotations': 'x/D 0.25/1440;',
            'data': [{'c': 59.61,
                      'h': 59.94,
                      'l': 51.86,
                      'o': 52.06,
                      't': 1698796800000,
                      'v': 1299913.92}],
            'high': '5994/1299913.92/1440',
            'low': '5186/1299913.92/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195422',
            'startTime': '20231031-00:00:00',
            'symbol': 'SCHW',
            'text': 'SCHWAB (CHARLES) CORP',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        },
        'TEAM': {
            'barLength': 2678400,
            'data': [{'c': 195.55,
                      'h': 199.36,
                      'l': 165.51,
                      'o': 180.7,
                      't': 1698796800000,
                      'v': 171177.85}],
            'high': '19936/171177.85/1440',
            'low': '16551/171177.85/1440',
            'mdAvailability': 'S',
            'messageVersion': 2,
            'mktDataDelay': 0,
            'negativeCapable': False,
            'outsideRth': False,
            'points': 0,
            'priceDisplayRule': 1,
            'priceDisplayValue': '2',
            'priceFactor': 100,
            'serverId': '195426',
            'startTime': '20231031-00:00:00',
            'symbol': 'TEAM',
            'text': 'ATLASSIAN CORP-CL A',
            'timePeriod': '1m',
            'travelTime': 2,
            'volumeFactor': 1
        }
    },
    "ledger": {
        "BASE": {
            "acctcode": "DUMMY_ACCOUNT_ID123",
            "cashbalance": 985376.94,
            "cashbalancefxsegment": 0.0,
            "commoditymarketvalue": 0.0,
            "corporatebondsmarketvalue": 0.0,
            "cryptocurrencyvalue": 0.0,
            "currency": "BASE",
            "dividends": 0.0,
            "exchangerate": 1,
            "funds": 0.0,
            "futuremarketvalue": 0.0,
            "futureoptionmarketvalue": 0.0,
            "futuresonlypnl": 0.0,
            "interest": 3401.28,
            "issueroptionsmarketvalue": 0.0,
            "key": "LedgerList",
            "moneyfunds": 0.0,
            "netliquidationvalue": 1034222.25,
            "realizedpnl": 0.0,
            "secondkey": "BASE",
            "sessionid": 1,
            "settledcash": 985376.94,
            "severity": 0,
            "stockmarketvalue": 45444.0,
            "stockoptionmarketvalue": 0.0,
            "tbillsmarketvalue": 0.0,
            "tbondsmarketvalue": 0.0,
            "timestamp": 1701198108,
            "unrealizedpnl": 1701.0,
            "warrantsmarketvalue": 0.0,
        },
        "USD": {
            "acctcode": "DUMMY_ACCOUNT_ID123",
            "cashbalance": 985376.94,
            "cashbalancefxsegment": 0.0,
            "commoditymarketvalue": 0.0,
            "corporatebondsmarketvalue": 0.0,
            "cryptocurrencyvalue": 0.0,
            "currency": "USD",
            "dividends": 0.0,
            "endofbundle": 1,
            "exchangerate": 1,
            "funds": 0.0,
            "futuremarketvalue": 0.0,
            "futureoptionmarketvalue": 0.0,
            "futuresonlypnl": 0.0,
            "interest": 3401.28,
            "issueroptionsmarketvalue": 0.0,
            "key": "LedgerList",
            "moneyfunds": 0.0,
            "netliquidationvalue": 1034222.25,
            "realizedpnl": 0.0,
            "secondkey": "USD",
            "sessionid": 1,
            "settledcash": 985376.94,
            "severity": 0,
            "stockmarketvalue": 45444.0,
            "stockoptionmarketvalue": 0.0,
            "tbillsmarketvalue": 0.0,
            "tbondsmarketvalue": 0.0,
            "timestamp": 1701198108,
            "unrealizedpnl": 1701.0,
            "warrantsmarketvalue": 0.0,
        },
    },
    "positions": [
        {
            "acctId": "DUMMY_ACCOUNT_ID123",
            "allExchanges": "AMEX,NYSE,CBOE,PHLX,CHX,ARCA,ISLAND,ISE,IDEAL,NASDAQQ,DRCTEDGE,BEX,BATS,NITEECN,EDGEA,CSFBALGO,JEFFALGO,PSX,BYX,ITG,PDQ,IBKRATS,CITADEL,NYSEDARK,MIAX,IBDARK,CITADELDP,NASDDARK,IEX,WEDBUSH,SUMMER,WINSLOW,LIQITG,UBSDARK,BTIG,VIRTU,JEFF,OPCO,COWEN,DBK,JPMC,EDGX,JANE,NEEDHAM,FRACSHARE,RBCALGO,VIRTUDP,BAYCREST,FOXRIVER,MND,NITEEXST,PEARL,GSDARK,NITERTL,NYSENAT,IEXMID,HRT,FLOWTRADE,HRTDP,JANELP,PEAK6,IMCDP,CTDLZERO,HRTMID,JANEZERO,HRTEXST,IMCLP,LTSE,SOCGENDP,MEMX,INTELCROS,VIRTUBYIN,JUMPTRADE,NITEZERO,TPLUS1,XTXEXST,XTXDP,XTXMID,COWENLP,BARCDP,JUMPLP,OLDMCLP,RBCCMALP,WALLBETH,IBEOS,JONES,GSLP,BLUEOCEAN,USIBSILP,OVERNIGHT,JANEMID,IBATSEOS,HRTZERO",
            "assetClass": "STK",
            "avgCost": 437.43,
            "avgPrice": 437.43,
            "chineseName": "SPDR&#x6807;&#x666E;500 ETF&#x4FE1;&#x6258;",
            "conExchMap": [],
            "conid": 756733,
            "contractDesc": "SPY",
            "countryCode": "US",
            "currency": "USD",
            "displayRule": {
                "displayRuleStep": [
                    {"decimalDigits": 2, "lowerEdge": 0.0, "wholeDigits": 4}
                ],
                "magnification": 0,
            },
            "exchs": None,
            "exerciseStyle": None,
            "expiry": None,
            "fullName": "SPY",
            "group": None,
            "hasOptions": True,
            "incrementRules": [{"increment": 0.01, "lowerEdge": 0.0}],
            "isEventContract": False,
            "isUS": True,
            "lastTradingDay": None,
            "listingExchange": "ARCA",
            "mktPrice": 454.44000245,
            "mktValue": 45444.0,
            "model": "",
            "multiplier": 0.0,
            "name": "SPDR S&P 500 ETF TRUST",
            "pageSize": 100,
            "position": 100.0,
            "putOrCall": None,
            "realizedPnl": 0.0,
            "sector": None,
            "sectorGroup": None,
            "strike": "0",
            "ticker": "SPY",
            "time": 30,
            "type": "ETF",
            "undConid": 0,
            "unrealizedPnl": 1701.0,
        }
    ],
    "stocks": {
        "AAPL": [
            {
                "assetClass": "STK",
                "chineseName": "&#x82F9;&#x679C;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 265598, "exchange": "NASDAQ", "isUS": True},
                    {"conid": 38708077, "exchange": "MEXI", "isUS": False},
                    {"conid": 273982664, "exchange": "EBS", "isUS": False},
                ],
                "name": "APPLE INC",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [
                    {"conid": 493546048, "exchange": "LSEETF", "isUS": False}
                ],
                "name": "LS 1X AAPL",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x82F9;&#x679C;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 532640894, "exchange": "AEQLIT", "isUS": False}
                ],
                "name": "APPLE INC-CDR",
            },
        ],
        "BBVA": [
            {
                "assetClass": "STK",
                "chineseName": "&#x897F;&#x73ED;&#x7259;&#x5BF9;&#x5916;&#x94F6;&#x884C;",
                "contracts": [
                    {"conid": 30314149, "exchange": "BM", "isUS": False},
                    {"conid": 38708792, "exchange": "MEXI", "isUS": False},
                ],
                "name": "BANCO BILBAO VIZCAYA ARGENTA",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x897F;&#x73ED;&#x7259;&#x5BF9;&#x5916;&#x94F6;&#x884C;",
                "contracts": [{"conid": 4815, "exchange": "NYSE", "isUS": True}],
                "name": "BANCO BILBAO VIZCAYA-SP ADR",
            },
        ],
        "CDN": [
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 453310666, "exchange": "PURE", "isUS": False}],
                "name": "CDN MAVERICK CAPITAL CORP",
            }
        ],
        "CFC": [
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 42001300, "exchange": "IBIS", "isUS": False}],
                "name": "UET UNITED ELECTRONIC TECHNO",
            }
        ],
        "GOOG": [
            {
                "assetClass": "STK",
                "chineseName": "Alphabet&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 208813720, "exchange": "NASDAQ", "isUS": True},
                    {"conid": 210810667, "exchange": "MEXI", "isUS": False},
                ],
                "name": "ALPHABET INC-CL C",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [
                    {"conid": 493546022, "exchange": "LSEETF", "isUS": False}
                ],
                "name": "LS 1X GOOG",
            },
            {
                "assetClass": "STK",
                "chineseName": "Alphabet&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 532638805, "exchange": "AEQLIT", "isUS": False}
                ],
                "name": "ALPHABET INC - CDR",
            },
        ],
        "HUBS": [
            {
                "assetClass": "STK",
                "chineseName": "HubSpot&#x516C;&#x53F8;",
                "contracts": [{"conid": 169544810, "exchange": "NYSE", "isUS": True}],
                "name": "HUBSPOT INC",
            }
        ],
        "META": [
            {
                "assetClass": "STK",
                "chineseName": "Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 107113386, "exchange": "NASDAQ", "isUS": True},
                    {"conid": 114922621, "exchange": "MEXI", "isUS": False},
                ],
                "name": "META PLATFORMS INC-CLASS A",
            },
            {
                "assetClass": "STK",
                "chineseName": "WisdomTree&#x5DE5;&#x4E1A;&#x91D1;&#x5C5E;&#x589E;&#x5F3A;",
                "contracts": [
                    {"conid": 355191857, "exchange": "LSEETF", "isUS": False},
                    {"conid": 355194893, "exchange": "BVME.ETF", "isUS": False},
                ],
                "name": "WT INDUS METALS ENHANCED",
            },
            {
                "assetClass": "STK",
                "chineseName": "Meta&#x5E73;&#x53F0;&#x80A1;&#x4EFD;&#x6709;&#x9650;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 530091499, "exchange": "AEQLIT", "isUS": False}
                ],
                "name": "META PLATFORMS INC-CDR",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 600541860, "exchange": "SFB", "isUS": False}],
                "name": "METACON AB",
            },
        ],
        "MSFT": [
            {
                "assetClass": "STK",
                "chineseName": "&#x5FAE;&#x8F6F;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 272093, "exchange": "NASDAQ", "isUS": True},
                    {"conid": 38708990, "exchange": "MEXI", "isUS": False},
                    {"conid": 415569505, "exchange": "EBS", "isUS": False},
                ],
                "name": "MICROSOFT CORP",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [
                    {"conid": 493546075, "exchange": "LSEETF", "isUS": False}
                ],
                "name": "LS 1X MSFT",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x5FAE;&#x8F6F;&#x516C;&#x53F8;",
                "contracts": [
                    {"conid": 518938052, "exchange": "AEQLIT", "isUS": False}
                ],
                "name": "MICROSOFT CORP-CDR",
            },
        ],
        "SAN": [
            {
                "assetClass": "STK",
                "chineseName": "&#x6851;&#x5766;&#x5FB7;",
                "contracts": [
                    {"conid": 38708867, "exchange": "MEXI", "isUS": False},
                    {"conid": 385055564, "exchange": "WSE", "isUS": False},
                ],
                "name": "BANCO SANTANDER SA",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x8D5B;&#x8BFA;&#x83F2;",
                "contracts": [{"conid": 29612249, "exchange": "SBF", "isUS": False}],
                "name": "SANOFI",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x6851;&#x5766;&#x5FB7;",
                "contracts": [{"conid": 12442, "exchange": "NYSE", "isUS": True}],
                "name": "BANCO SANTANDER SA-SPON ADR",
            },
            {
                "assetClass": "STK",
                "chineseName": "&#x6851;&#x5766;&#x5FB7;&#x82F1;&#x56FD;&#x516C;&#x5171;&#x6709;&#x9650;&#x516C;&#x53F8;",
                "contracts": [{"conid": 80993135, "exchange": "LSE", "isUS": False}],
                "name": "SANTANDER UK PLC",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 199284645, "exchange": "ASX", "isUS": False}],
                "name": "SAGALIO ENERGY LTD",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 221033380, "exchange": "VALUE", "isUS": False}],
                "name": "CONCORD GREEN ENERGY HOLDINGS",
            },
        ],
        "SCHW": [
            {
                "assetClass": "STK",
                "chineseName": "&#x5609;&#x4FE1;&#x7406;&#x8D22;",
                "contracts": [
                    {"conid": 11905, "exchange": "NYSE", "isUS": True},
                    {"conid": 174320486, "exchange": "MEXI", "isUS": False},
                ],
                "name": "SCHWAB (CHARLES) CORP",
            }
        ],
        "TEAM": [
            {
                "assetClass": "STK",
                "chineseName": "Team "
                               "Tankers&#x56FD;&#x9645;&#x6709;&#x9650;&#x516C;&#x53F8;",
                "contracts": [{"conid": 251279169, "exchange": "VALUE", "isUS": False}],
                "name": "TEAM TANKERS INTERNATIONAL L",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 475655801, "exchange": "LSE", "isUS": False}],
                "name": "TEAM PLC",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 589316251, "exchange": "NASDAQ", "isUS": True}],
                "name": "ATLASSIAN CORP-CL A",
            },
            {
                "assetClass": "STK",
                "chineseName": None,
                "contracts": [{"conid": 638360210, "exchange": "PURE", "isUS": False}],
                "name": "CANADIAN NEXUS TEAM VENTURES",
            },
        ],
    },
    "tickle": {
        "collission": False,
        "hmds": {"error": "no bridge"},
        "iserver": {
            "authStatus": {
                "MAC": "F4:03:43:CD:BD:C0",
                "authenticated": True,
                "competing": False,
                "connected": True,
                "hardware_info": "1c4558e8|F4:03:43:CD:BD:C0",
                "message": "",
                "serverInfo": {
                    "serverName": "JifC11113",
                    "serverVersion": "Build "
                                     "10.25.0m, "
                                     "Nov "
                                     "17, "
                                     "2023 "
                                     "11:29:27 "
                                     "AM",
                },
            }
        },
        "session": "ce109cbeb462a76dc3f6f4182c830cc1",
        "ssoExpires": 585695,
        "userId": 99996016,
    },
    "live_orders": [
        {'account': 'DUMMY_ACCOUNT_ID123',
         'acct': 'DUMMY_ACCOUNT_ID123',
         'bgColor': '#FFFFFF',
         'cashCcy': 'USD',
         'companyName': 'SPDR S&P 500 ETF TRUST',
         'conid': 756733,
         'conidex': '756733',
         'description1': 'SPY',
         'fgColor': '#0000CC',
         'filledQuantity': 0.0,
         'lastExecutionTime': '240103201041',
         'lastExecutionTime_r': 1704312641000,
         'listingExchange': 'ARCA',
         'orderDesc': 'Buy 1 Limit 457.00, GTC',
         'orderId': 1210386759,
         'orderType': 'Limit',
         'order_ccp_status': 'Submitted',
         'order_ref': '240103-SPY-1-BUY-L@457-bxv8',
         'origOrderType': 'LIMIT',
         'outsideRTH': False,
         'price': '457.00',
         'remainingQuantity': 1.0,
         'secType': 'STK',
         'side': 'BUY',
         'sizeAndFills': '0/1',
         'status': 'PreSubmitted',
         'supportsTaxOpt': '1',
         'ticker': 'SPY',
         'timeInForce': 'GTC',
         'totalSize': 1.0},
        {'account': 'DUMMY_ACCOUNT_ID123',
         'acct': 'DUMMY_ACCOUNT_ID123',
         'bgColor': '#FFFFFF',
         'cashCcy': 'USD',
         'companyName': 'SPDR S&P 500 ETF TRUST',
         'conid': 756733,
         'conidex': '756733',
         'description1': 'SPY',
         'fgColor': '#0000CC',
         'filledQuantity': 0.0,
         'lastExecutionTime': '240103200639',
         'lastExecutionTime_r': 1704312399000,
         'listingExchange': 'ARCA',
         'orderDesc': 'Buy 1 Limit 457.00, GTC',
         'orderId': 1210386726,
         'orderType': 'Limit',
         'order_ccp_status': 'Submitted',
         'order_ref': '240103-SPY-1-BUY-L@457-pfqd',
         'origOrderType': 'LIMIT',
         'outsideRTH': False,
         'price': '457.00',
         'remainingQuantity': 1.0,
         'secType': 'STK',
         'side': 'BUY',
         'sizeAndFills': '0/1',
         'status': 'PreSubmitted',
         'supportsTaxOpt': '1',
         'ticker': 'SPY',
         'timeInForce': 'GTC',
         'totalSize': 1.0},
        {'account': 'DUMMY_ACCOUNT_ID123',
         'acct': 'DUMMY_ACCOUNT_ID123',
         'bgColor': '#FFFFFF',
         'cashCcy': 'USD',
         'companyName': 'SPDR S&P 500 ETF TRUST',
         'conid': 756733,
         'conidex': '756733',
         'description1': 'SPY',
         'fgColor': '#0000CC',
         'filledQuantity': 0.0,
         'lastExecutionTime': '240103010227',
         'lastExecutionTime_r': 1704243747000,
         'listingExchange': 'ARCA',
         'orderDesc': 'Buy 1 Limit 453.00, GTC',
         'orderId': 1928088808,
         'orderType': 'Limit',
         'order_ccp_status': 'Submitted',
         'order_ref': '240103-SPY-1-BUY-L@453-v5DU',
         'origOrderType': 'LIMIT',
         'outsideRTH': False,
         'price': '453.00',
         'remainingQuantity': 1.0,
         'secType': 'STK',
         'side': 'BUY',
         'sizeAndFills': '0/1',
         'status': 'PreSubmitted',
         'supportsTaxOpt': '1',
         'ticker': 'SPY',
         'timeInForce': 'GTC',
         'totalSize': 1.0},
        {'account': 'DUMMY_ACCOUNT_ID123',
         'acct': 'DUMMY_ACCOUNT_ID123',
         'bgColor': '#FFFFFF',
         'cashCcy': 'USD',
         'companyName': 'SPDR S&P 500 ETF TRUST',
         'conid': 756733,
         'conidex': '756733',
         'description1': 'SPY',
         'fgColor': '#0000CC',
         'filledQuantity': 0.0,
         'lastExecutionTime': '240103212831',
         'lastExecutionTime_r': 1704317311000,
         'listingExchange': 'ARCA',
         'orderDesc': 'Buy 1 Limit 458.00, GTC',
         'orderId': 1210387040,
         'orderType': 'Limit',
         'order_ccp_status': 'Submitted',
         'order_ref': '240103-SPY-1-BUY-L@458-J5cN',
         'origOrderType': 'LIMIT',
         'outsideRTH': False,
         'price': '458.00',
         'remainingQuantity': 1.0,
         'secType': 'STK',
         'side': 'BUY',
         'sizeAndFills': '0/1',
         'status': 'PreSubmitted',
         'supportsTaxOpt': '1',
         'ticker': 'SPY',
         'timeInForce': 'GTC',
         'totalSize': 1.0}
    ],
    "order_status": {
        '1210386759': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386759,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103201041',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37888',
            'sec_type': 'STK',
            'server_id': '0',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False},
        '1210386726': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386726,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103200639',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37889',
            'sec_type': 'STK',
            'server_id': '1',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False},
        '1928088808': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#0000CC',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 453.00, GTC',
            'exit_strategy_display_price': '453.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#FFFFFF',
            'limit_price': '453.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 453.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 453.00, GTC',
            'order_id': 1928088808,
            'order_not_editable': False,
            'order_status': 'PreSubmitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103010227',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '37890',
            'sec_type': 'STK',
            'server_id': '2',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False},
        '1210386714': {
            'account': 'DUMMY_ACCOUNT_ID123',
            'alert_active': 1,
            'allowed_duplicate_opposite': True,
            'available_chart_periods': '#R|1',
            'bg_color': '#00F000',
            'cannot_cancel_order': False,
            'child_order_type': '3',
            'company_name': 'SPDR S&P 500 ETF TRUST',
            'conid': 756733,
            'conidex': '756733',
            'contract_description_1': 'SPY',
            'cum_fill': '0.0',
            'currency': 'USD',
            'deactivate_order': False,
            'editable_fields': '\x1e',
            'exit_strategy_chart_description': 'Buy 1 Limit 457.00, GTC',
            'exit_strategy_display_price': '457.00',
            'exit_strategy_tool_availability': '1',
            'fg_color': '#000000',
            'limit_price': '457.00',
            'listing_exchange': 'ARCA',
            'option_acct': 'c',
            'order_ccp_status': '0',
            'order_clearing_account': 'DUMMY_ACCOUNT_ID123',
            'order_description': 'Buy 1 Limit 457.00, GTC',
            'order_description_with_contract': 'Buy 1 SPY Limit 457.00, GTC',
            'order_id': 1210386714,
            'order_not_editable': False,
            'order_status': 'Submitted',
            'order_status_description': 'Order Submitted',
            'order_time': '240103200503',
            'order_type': 'LIMIT',
            'outside_rth': False,
            'request_id': '216243',
            'sec_type': 'STK',
            'server_id': '824',
            'side': 'B',
            'size': '1.0',
            'size_and_fills': '0/1',
            'sub_type': None,
            'symbol': 'SPY',
            'tif': 'GTC',
            'total_size': '1.0',
            'use_price_mgmt_algo': False}
    },

}



import datetime
import logging
import sys
from pathlib import Path

from ibind import var

DEFAULT_FORMAT = '%(asctime)s|%(levelname)-.1s| %(message)s'

_initialized = False
_log_to_file = False


def project_logger(filepath=None):
    """
    Returns a project-specific logger instance.

    This function creates a new logger instance with the name `'ibind'.{filepath.stem}` if `filepath` is specified,
    or 'ibind' if `filepath` is not specified.

    Parameters:
        filepath (str): The file path where the log files will be stored. If no filepath is specified, the default logs directory will be used.

    Returns:
        logging.Logger: The project-specific logger instance.
    """
    return logging.getLogger('ibind' + (f'.{Path(filepath).stem}' if filepath is not None else ''))

_LOGGER = project_logger()

def ibind_logs_initialize(
        log_to_console: bool = var.LOG_TO_CONSOLE,
        log_to_file: bool = var.LOG_TO_FILE,
        log_level: str = var.LOG_LEVEL,
        log_format: str = var.LOG_FORMAT,
):
    """
    Initialises the logging system.

    Parameters:
        log_to_console (bool): Whether the logs should be output to the current console, `True` by default
        log_to_file (bool): Whether the logs should be written to a daily log file, `True` by default.
        log_level (str): What is the minimum log level of `ibind` logs, `INFO` by default.
        log_format (str): What is the log format to be used, `'%(asctime)s|%(levelname)-.1s| %(message)s'` by default.

    Note:
        - All of these parameters are read from the environment variables by default.
        - The daily file logs are saved in the directory specified by the `IBIND_LOGS_DIR` environment variable, the system temp directory by default.
        - To get more verbose logs, set either the `log_level` parameter or the `IBIND_LOG_LEVEL` environment variable to `'DEBUG'`
    """
    global _initialized
    if _initialized: return
    _initialized = True

    global _log_to_file
    _log_to_file = log_to_file


    logger = logging.getLogger('ibind')
    formatter = logging.Formatter(log_format)
    logger.setLevel(logging.DEBUG)

    if log_to_console:
        # outputting only to a single stream to ensure chronological ordering of all messages
        h1 = logging.StreamHandler(stream=sys.stdout)
        h1.setLevel(getattr(logging, log_level))
        h1.setFormatter(formatter)
        logger.addHandler(h1)

    if not _log_to_file:
        logging.getLogger('ibind_fh').addFilter(lambda record: False)


def new_daily_rotating_file_handler(logger_name, filepath):
    """
    Creates and configures a new daily rotating file handler for logging.

    This function sets up a file-based logger with a daily rotation. It ensures that each day's logs are
    saved in a separate file. The logger is configured with a specific name and file path. If the logger
    with the given name doesn't already have handlers, it adds a new daily rotating file handler to it.

    Parameters:
        logger_name (str): The name to be assigned to the logger. This name is used to identify the logger instance.
        filepath (str): The file path where the log files will be stored. This path determines the location of the log files.

    Returns:
        logging.Logger: The logger configured with the daily rotating file handler.

    Note:
        - The logger is set to DEBUG level by default.
        - The format of the logs is determined by the DEFAULT_FORMAT global variable.
    """
    logger = logging.getLogger(f'ibind_fh.{logger_name}')

    if _log_to_file:
        _LOGGER.info(f'New daily rotating file handler for logger "{logger_name}": {filepath}')
        if len(logger.handlers) == 0:
            fh_logger = logging.getLogger('ibind_fh')
            handler = DailyRotatingFileHandler(filepath, encoding='utf-8')
            handler.setFormatter(logging.Formatter(DEFAULT_FORMAT))

            # if filehandler outputs are disabled, this should bring over the filter that will do this
            for filter in fh_logger.filters:
                logger.addFilter(filter)
            logger.addHandler(handler)

        logger.setLevel(logging.DEBUG)
    else:
        logger.addHandler(logging.NullHandler())

    return logger


class DailyRotatingFileHandler(logging.FileHandler):

    def __init__(self, *args, date_format='%Y-%m-%d', **kwargs):
        self.timestamp = None
        self.date_format = date_format
        self.stream = None
        super().__init__(*args, **kwargs)

    def get_timestamp(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        return now.strftime(self.date_format)

    def get_filename(self, timestamp):
        return f'{self.baseFilename}__{timestamp}.txt'

    def _open(self):
        if self.stream is not None:
            self.close()
        self.timestamp = self.get_timestamp()
        filename = self.get_filename(self.timestamp)
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        return open(filename, self.mode, encoding='utf-8')

    def emit(self, record):
        if self.get_timestamp() != self.timestamp:
            self.close()
            self.stream = self._open()

        super().emit(record)




import datetime
from typing import Union, TYPE_CHECKING, List

from ibind.base.rest_client import Result
from ibind.client.ibkr_definitions import decode_data_availability, snapshot_by_id
from ibind.client.ibkr_utils import StockQuery, StockQueries
from ibind.support.errors import ExternalBrokerError
from ibind.support.logs import project_logger
from ibind.support.py_utils import ensure_list_arg, OneOrMany, execute_in_parallel, params_dict

if TYPE_CHECKING:  # pragma: no cover
    from ibind import IbkrClient

_LOGGER = project_logger(__file__)


class MarketdataMixin():
    """
    https://ibkrcampus.com/ibkr-api-page/cpapi-v1/#md
    """

    @ensure_list_arg('conids', 'fields')
    def live_marketdata_snapshot(self: 'IbkrClient', conids: OneOrMany[str], fields: OneOrMany[str]) -> Result:  # pragma: no cover
        """
        Get Market Data for the given conid(s).

        A pre-flight request must be made prior to ever receiving data.

        Parameters:
            conids (OneOrMany[str]): Contract identifier(s) for the contract of interest.
            fields (OneOrMany[str]): Specify a series of tick values to be returned.

        Note:
            - The endpoint /iserver/accounts must be called prior to /iserver/marketdata/snapshot.
            - For derivative contracts, the endpoint /iserver/secdef/search must be called first.
        """
        params = {
            'conids': ','.join(conids),
            'fields': ','.join(fields)
        }
        return self.get(f'iserver/marketdata/snapshot', params)

    def live_marketdata_snapshot_by_symbol(self: 'IbkrClient', queries: StockQueries, fields: OneOrMany[str]) -> dict:
        """
        Get Market Data for the given symbols(s).

        A pre-flight request must be made prior to ever receiving data.

        Parameters:
            queries (List[StockQuery]): A list of StockQuery objects to specify filtering criteria for stocks.
            fields (OneOrMany[str]): Specify a series of tick values to be returned.

        Note:
            - The endpoint /iserver/accounts must be called prior to /iserver/marketdata/snapshot.
            - For derivative contracts, the endpoint /iserver/secdef/search must be called first.
        """
        conids_by_symbol = self.stock_conid_by_symbol(queries).data

        conids = []
        symbols_by_conids = {}
        for symbol, conid in conids_by_symbol.items():
            conids.append(str(conid))
            symbols_by_conids[conid] = symbol

        # There needs to be a pre-flight request for some reason, so we default to calling this twice
        self.receive_brokerage_accounts()
        self.live_marketdata_snapshot(conids, fields=fields)
        entries = self.live_marketdata_snapshot(conids, fields=fields).data

        results = {}
        for entry in entries:
            try:
                result = {}

                for key, value in entry.items():
                    if key not in snapshot_by_id:
                        continue

                    result[snapshot_by_id[key]] = value
                results[entry['conid']] = result
            except Exception as e:  # pragma: no cover
                _LOGGER.exception(f'Error post-processing live market data for {entry}: {str(e)}')

        # reformat the results by symbol instead of conid
        results_by_symbol = {symbols_by_conids[conid]: result for conid, result in results.items()}

        return results_by_symbol

    def regulatory_snapshot(self: 'IbkrClient', conid: str) -> Result:  # pragma: no cover
        """
        Send a request for a regulatory snapshot. This will cost $0.01 USD per request unless you are subscribed to the direct exchange market data already.

        WARNING: Each regulatory snapshot made will incur a fee of $0.01 USD to the account. This applies to both live and paper accounts.

        Parameters:
            conid (str): Provide the contract identifier to retrieve market data for.

        Note:
            - If you are already paying for, or are subscribed to, a specific US Network subscription, your account will not be charged.
            - For stocks, there are individual exchange-specific market data subscriptions necessary to receive streaming quotes.
        """
        return self.get(f'md/regsnapshot', {'conid': conid})

    def marketdata_history_by_conid(
            self: 'IbkrClient',
            conid: str,
            bar: str,
            exchange: str = None,
            period: str = None,
            outside_rth: bool = None,
            start_time: datetime.datetime = None
    ) -> Result:  # pragma: no cover
        """
        Get historical market Data for given conid, length of data is controlled by 'period' and 'bar'.

        Parameters:
            conid (str): Contract identifier for the ticker symbol of interest.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.

        Note:
            - There's a limit of 5 concurrent requests. Excessive requests will return a 'Too many requests' status 429 response.
        """
        params = params_dict(
            {
                'conid': conid,
                'bar': bar
            },
            optional={
                'exchange': exchange,
                'period': period,
                'outsideRth': outside_rth,
                'startTime': start_time
            },
            preprocessors={
                'startTime': lambda x: x.strftime('%Y%m%d-%H:%M:%S')
            }
        )

        return self.get('iserver/marketdata/history', params)

    def historical_marketdata_beta(
            self: 'IbkrClient',
            conid: str,
            period: str,
            bar: str,
            outside_rth: bool = None,
            start_time: datetime.datetime = None,
            direction: str = None,
            bar_type: str = None,
    ) -> Result:  # pragma: no cover
        """
        Using a direct connection to the market data farm, will provide a list of historical market data for given conid.

        Parameters:
            conid (str): The contract identifier for which data should be requested.
            period (str): The duration for which data should be requested. Available Values: See HMDS Period Units.
            bar (str): The bar size for which bars should be returned. Available Values: See HMDS Bar Sizes.
            outside_rth (bool, optional): Define if data should be returned for trades outside regular trading hours.
            start_time (datetime.datetime, optional): Specify the value from where historical data should be taken. Value Format: UTC; YYYYMMDD-HH:mm:dd. Defaults to the current date and time.
            direction (str, optional): Specify the direction from which market data should be returned. Available Values: -1: time from the start_time to now; 1: time from now to the end of the period. Defaults to 1.
            bar_type (str, optional): Returns valid bar types for which data may be requested. Available Values: Last, Bid, Ask, Midpoint, FeeRate, Inventory. Defaults to Last for Stocks, Options, Futures, and Futures Options.

        Note:
            - The first time a user makes a request to the /hmds/history endpoints will result in a 404 error. This initial request instantiates the historical market data services allowing future requests to return data. Subsequent requests will return data as expected.
        """
        params = params_dict(
            {
                'conid': conid,
                'period': period,
                'bar': bar
            },
            optional={
                'outsideRth': outside_rth,
                'startTime': start_time,
                'direction': direction,
                'barType': bar_type,
            },
            preprocessors={
                'startTime': lambda x: x.strftime('%Y%m%d-%H:%M:%S')
            }
        )

        return self.get('hmds/history', params)

    def marketdata_history_by_symbol(
            self: 'IbkrClient',
            symbol: Union[str, StockQuery],
            bar: str,
            exchange: str = None,
            period: str = None,
            outside_rth: bool = None,
            start_time: datetime.datetime = None,
    ) -> Result:  # pragma: no cover
        """
        Get historical market Data for given symbol, length of data is controlled by 'period' and 'bar'.

        Parameters:
            symbol (Union[str, StockQuery]): StockQuery or str symbol for the ticker of interest.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.

        """
        conid = str(self.stock_conid_by_symbol(symbol).data[symbol])
        return self.marketdata_history_by_conid(conid, bar, exchange, period, outside_rth, start_time)

    @ensure_list_arg('queries')
    def marketdata_history_by_symbols(
            self: 'IbkrClient',
            queries: StockQueries,
            period: str = "1min",
            bar: str = "1min",
            outside_rth: bool = True,
            start_time: datetime.datetime = None,
    ) -> dict:
        """
        An extended version of the marketdata_history_by_symbol method.

        For each StockQuery provided, it queries the marketdata history for the specified symbols in parallel. The results are then cleaned up and unified. Due to this grouping and post-processing, this method returns data directly without the Result dataclass.

        Parameters:
            queries (List[StockQuery]): A list of StockQuery objects to specify filtering criteria for stocks.
            exchange (str, optional): Returns the exchange you want to receive data from.
            period (str): Overall duration for which data should be returned. Default to 1w. Available time period– {1-30}min, {1-8}h, {1-1000}d, {1-792}w, {1-182}m, {1-15}y.
            bar (str): Individual bars of data to be returned. Possible values– 1min, 2min, 3min, 5min, 10min, 15min, 30min, 1h, 2h, 3h, 4h, 8h, 1d, 1w, 1m.
            outside_rth (bool, optional): Determine if you want data after regular trading hours.
            start_time (datetime.datetime, optional): Starting date of the request duration.

        Note:
            - This method returns data directly without the `Result` dataclass.
        """
        conids = self.stock_conid_by_symbol(queries).data

        static_params = {"period": period, "bar": bar, "outside_rth": outside_rth, 'start_time': start_time}
        requests = {symbol: {"kwargs": {'conid': conid} | static_params} for symbol, conid in conids.items()}

        # /iserver/marketdata/history accepts 5 concurrent requests at a time
        history = execute_in_parallel(self.marketdata_history_by_conid, requests=requests, max_workers=5)

        results = {}
        for symbol, entry in history.items():
            if isinstance(entry, Exception):  # pragma: no cover
                _LOGGER.error(f'Error fetching market data for {symbol}')
                raise entry

            # check if entry['mdAvailability'] has 'S' or 'R' in it
            if 'mdAvailability' in entry.data and not (any((key in entry.data['mdAvailability'].upper()) for key in ['S', 'R'])):
                _LOGGER.warning(f'Market data for {symbol} is not live: {decode_data_availability(entry.data["mdAvailability"])}')

            data = entry.data['data']
            records = []
            for record in data:
                records.append({
                    "open": record['o'],
                    "high": record['h'],
                    "low": record['l'],
                    "close": record['c'],
                    "volume": record['v'],
                    "date": datetime.datetime.fromtimestamp(record['t'] / 1000)
                })
            results[symbol] = records

        return results

    @ensure_list_arg('conids')
    def marketdata_unsubscribe(self: 'IbkrClient', conids: OneOrMany[str]) -> List[Result]:
        """
        Cancel market data for given conid(s).

        Parameters:
            conids (OneOrMany[str]): Enter the contract identifier to cancel the market data feed. This can clear all standing market data feeds to invalidate your cache and start fresh.
        """
        # we unsubscribe from all conids simultaneously
        unsubscribe_requests = {conid: {'args': [f'iserver/marketdata/{conid}/unsubscribe']} for conid in conids}
        results = execute_in_parallel(self.post, unsubscribe_requests)

        for conid, result in results.items():
            if isinstance(result, Exception):
                # 404 means that no such subscription was found in first place, which we ignore
                if isinstance(result, ExternalBrokerError) and result.status_code == 404:
                    continue
                raise result

        return results

    def marketdata_unsubscribe_all(self: 'IbkrClient') -> Result:  # pragma: no cover
        """
        Cancel all market data request(s). To cancel market data for a specific conid, see /iserver/marketdata/{conid}/unsubscribe.
        """
        return self.get(f'iserver/marketdata/unsubscribeall')



from distutils.core import setup

from setuptools import find_packages


setup(
    name='ibind',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "examples", "docs", "out", "dist"])
)
